# Copyright 2022 macanudo527
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Autoinvest CSV Format: timestamp UTC, base asset symbol, quote asset amount + symbol, trading fee (in quote asset),
#    base asset amount + symbol, source of funds
# Note: file comes as .xlsx, and then needs to be saved as CSV.

# Betheth CSV format: timestamp UTC, quote asset symbol (ETH), base asset symbol (BETH), amount, status

import logging
from csv import reader
from typing import List, Optional

from rp2.abstract_country import AbstractCountry
from rp2.logger import create_logger
from rp2.rp2_decimal import RP2Decimal

from dali.abstract_input_plugin import AbstractInputPlugin
from dali.abstract_transaction import AbstractTransaction
from dali.configuration import Keyword
from dali.in_transaction import InTransaction
from dali.out_transaction import OutTransaction


class InputPlugin(AbstractInputPlugin):

    __BINANCE_COM: str = "Binance.com"
    __BINANCE_COM_SUPPLEMENTAL_CSV: str = "Binance.com_Supplemental_CSV"

    __TIMESTAMP_INDEX: int = 0
    __AUTO_BASE_SYMBOL: int = 1
    __AUTO_QUOTE_AMOUNT_SYMBOL: int = 2
    __AUTO_TRADING_FEE_SYMBOL: int = 3
    __BETHETH_AMOUNT: int = 3
    __AUTO_BASE_AMOUNT_SYMBOL: int = 4
    __FUND_SOURCE: int = 5

    __DELIMITER: str = ","

    def __init__(
        self,
        account_holder: str,
        autoinvest_csv_file: Optional[str] = None,
        betheth_csv_file: Optional[str] = None,
        native_fiat: Optional[str] = None,
    ) -> None:

        super().__init__(account_holder=account_holder, native_fiat=native_fiat)
        self.__autoinvest_csv_file: Optional[str] = autoinvest_csv_file
        self.__betheth_csv_file: Optional[str] = betheth_csv_file
        self.__logger: logging.Logger = create_logger(f"{self.__BINANCE_COM_SUPPLEMENTAL_CSV}/{self.account_holder}")

    def load(self, country: AbstractCountry) -> List[AbstractTransaction]:
        result: List[AbstractTransaction] = []

        if self.__autoinvest_csv_file:
            result.extend(self.parse_autoinvest_file(self.__autoinvest_csv_file))

        if self.__betheth_csv_file:
            result.extend(self.parse_betheth_file(self.__betheth_csv_file))

        self.__logger.debug("Binance_CSV results %s", result)

        return result

    def parse_autoinvest_file(self, file_path: str) -> List[AbstractTransaction]:
        result: List[AbstractTransaction] = []

        with open(file_path, encoding="utf-8") as csv_file:
            lines = reader(csv_file)

            header = next(lines)
            self.__logger.debug("Header: %s", header)
            for line in lines:
                raw_data: str = self.__DELIMITER.join(line)
                self.__logger.debug("Transaction: %s", raw_data)

                result.append(
                    InTransaction(
                        plugin=self.__BINANCE_COM_SUPPLEMENTAL_CSV,
                        unique_id=Keyword.UNKNOWN.value,
                        raw_data=raw_data,
                        timestamp=f"{line[self.__TIMESTAMP_INDEX]} -00:00",
                        asset=line[self.__AUTO_BASE_SYMBOL],
                        exchange=self.__BINANCE_COM,
                        holder=self.account_holder,
                        transaction_type=Keyword.BUY.value,
                        spot_price=Keyword.UNKNOWN.value,
                        crypto_in=(line[self.__AUTO_BASE_AMOUNT_SYMBOL]).split()[0],
                        notes=f"Autoinvest buy with funding from {line[self.__FUND_SOURCE]}",
                    )
                )

                quote_asset_symbol: str = line[self.__AUTO_QUOTE_AMOUNT_SYMBOL].split()[1]
                quote_asset_amount: str = line[self.__AUTO_QUOTE_AMOUNT_SYMBOL].split()[0]
                crypto_fee: str = line[self.__AUTO_TRADING_FEE_SYMBOL].split()[0]
                crypto_out_with_fee: RP2Decimal = RP2Decimal(quote_asset_amount) + RP2Decimal(crypto_fee)
                result.append(
                    OutTransaction(
                        plugin=self.__BINANCE_COM_SUPPLEMENTAL_CSV,
                        unique_id=Keyword.UNKNOWN.value,
                        raw_data=raw_data,
                        timestamp=f"{line[self.__TIMESTAMP_INDEX]} -00:00",
                        asset=quote_asset_symbol,
                        exchange=self.__BINANCE_COM,
                        holder=self.account_holder,
                        transaction_type=Keyword.SELL.value,
                        spot_price=Keyword.UNKNOWN.value,
                        crypto_out_no_fee=quote_asset_amount,
                        crypto_out_with_fee=str(crypto_out_with_fee),
                        crypto_fee=crypto_fee,
                        notes=f"Autoinvest withdrawal for the purchase of {line[self.__AUTO_BASE_AMOUNT_SYMBOL]}",
                    )
                )

        return result

    def parse_betheth_file(self, file_path: str) -> List[AbstractTransaction]:
        result: List[AbstractTransaction] = []

        with open(file_path, encoding="utf-8") as csv_file:
            lines = reader(csv_file)

            header = next(lines)
            self.__logger.debug("Header: %s", header)
            for line in lines:
                raw_data: str = self.__DELIMITER.join(line)
                self.__logger.debug("Transaction: %s", raw_data)

                result.append(
                    InTransaction(
                        plugin=self.__BINANCE_COM_SUPPLEMENTAL_CSV,
                        unique_id=Keyword.UNKNOWN.value,
                        raw_data=raw_data,
                        timestamp=f"{line[self.__TIMESTAMP_INDEX]} -00:00",
                        asset="BETH",
                        exchange=self.__BINANCE_COM,
                        holder=self.account_holder,
                        transaction_type=Keyword.BUY.value,
                        spot_price=Keyword.UNKNOWN.value,
                        crypto_in=line[self.__BETHETH_AMOUNT],
                        notes="Conversion from ETH -> BETH",
                    )
                )

                result.append(
                    OutTransaction(
                        plugin=self.__BINANCE_COM_SUPPLEMENTAL_CSV,
                        unique_id=Keyword.UNKNOWN.value,
                        raw_data=raw_data,
                        timestamp=f"{line[self.__TIMESTAMP_INDEX]} -00:00",
                        asset="ETH",
                        exchange=self.__BINANCE_COM,
                        holder=self.account_holder,
                        transaction_type=Keyword.SELL.value,
                        spot_price=Keyword.UNKNOWN.value,
                        crypto_out_no_fee=line[self.__BETHETH_AMOUNT],
                        crypto_out_with_fee=line[self.__BETHETH_AMOUNT],
                        crypto_fee="0",
                        notes="Conversion from ETH -> BETH",
                    )
                )

        return result



# Copyright 2022 Steve Davis
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import NamedTuple, Optional

import pytest

from dali.configuration import Keyword
from dali.intra_transaction import IntraTransaction
from dali.transaction_resolver import _resolve_intra_intra_transaction


class IntraTransactionNotesTestCase(NamedTuple):
    prior_notes: Optional[str]
    notes1: Optional[str]
    notes2: Optional[str]
    resolved_notes: Optional[str]


intra_transaction_notes_test_cases = [
    IntraTransactionNotesTestCase(None, None, None, None),  # resolver assignes empty string, IntraTransaction changes it to None
    IntraTransactionNotesTestCase(None, "xxx", None, "xxx;"),
    IntraTransactionNotesTestCase(None, None, "yyy", "yyy;"),
    IntraTransactionNotesTestCase(None, "xxx", "yyy", "xxx; yyy;"),
    IntraTransactionNotesTestCase(None, "xxx", "xxx", "xxx;"),  # do not concatentate identical notes
    IntraTransactionNotesTestCase("prior", None, None, "prior;"),
    IntraTransactionNotesTestCase("prior", "xxx", None, "prior; xxx;"),
    IntraTransactionNotesTestCase("prior", None, "yyy", "prior; yyy;"),
    IntraTransactionNotesTestCase("prior", "xxx", "yyy", "prior; xxx; yyy;"),
    IntraTransactionNotesTestCase("prior", "xxx", "xxx", "prior; xxx;"),  # do not concatentate identical notes
]


@pytest.mark.parametrize("prior_notes, notes1, notes2, resolved_notes", intra_transaction_notes_test_cases)
def test_resolve_intra_intra_transaction(prior_notes: Optional[str], notes1: Optional[str], notes2: Optional[str], resolved_notes: Optional[str]) -> None:
    """Verify resolved transaction matches expected values."""
    transaction1 = IntraTransaction(
        plugin="plugin",
        unique_id="unique_id",
        raw_data="raw_data1",
        asset="asset",
        timestamp="2022-01-01 00:00:00+00:00",
        from_exchange="from_exchange1",
        from_holder="from_holder1",
        to_exchange=Keyword.UNKNOWN.value,
        to_holder=Keyword.UNKNOWN.value,
        spot_price=None,
        crypto_sent="1.0",
        crypto_received="1.0",
        notes=notes1,
    )

    transaction2 = IntraTransaction(
        plugin="plugin",
        unique_id="unique_id",
        raw_data="raw_data2",
        asset="asset",
        timestamp="2022-01-01 00:00:00+00:00",
        from_exchange=Keyword.UNKNOWN.value,
        from_holder=Keyword.UNKNOWN.value,
        to_exchange="to_exchange2",
        to_holder="to_holder2",
        spot_price="1000.0",
        crypto_sent="1.0",
        crypto_received="1.0",
        notes=notes2,
    )

    resolved = _resolve_intra_intra_transaction(transaction1, transaction2, prior_notes)
    assert resolved.from_exchange == "from_exchange1"
    assert resolved.from_holder == "from_holder1"
    assert resolved.to_exchange == "to_exchange2"
    assert resolved.to_holder == "to_holder2"
    assert resolved.spot_price == "1000.0"
    assert resolved.notes == resolved_notes



# Copyright 2022 eprbell
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Coinbase REST plugin links:
# REST API: https://developers.coinbase.com/api/v2
# Authentication: https://developers.coinbase.com/docs/wallet/api-key-authentication
# Endpoint: https://api.coinbase.com

import hashlib
import hmac
import json
import logging
import time
from multiprocessing.pool import ThreadPool
from typing import Any, Dict, List, NamedTuple, Optional, cast

import requests
from requests import PreparedRequest
from requests.auth import AuthBase
from requests.models import Response
from requests.sessions import Session

from rp2.abstract_country import AbstractCountry
from rp2.logger import create_logger
from rp2.rp2_decimal import ZERO, RP2Decimal
from rp2.rp2_error import RP2RuntimeError

from dali.abstract_input_plugin import AbstractInputPlugin
from dali.abstract_transaction import AbstractTransaction
from dali.configuration import Keyword
from dali.in_transaction import InTransaction
from dali.intra_transaction import IntraTransaction
from dali.out_transaction import OutTransaction

# Native format keywords
_ADVANCED_TRADE_FILL: str = "advanced_trade_fill"
_AMOUNT: str = "amount"
_BALANCE: str = "balance"
_BUY: str = "buy"
_CARDBUYBACK: str = "cardbuyback"
_CARDSPEND: str = "cardspend"
_CODE: str = "code"
_COMMISSION: str = "commission"
_CREATED_AT: str = "created_at"
_CURRENCY: str = "currency"
_DETAILS: str = "details"
_EMAIL: str = "email"
_EXCHANGE_DEPOSIT: str = "exchange_deposit"
_EXCHANGE_WITHDRAWAL: str = "exchange_withdrawal"
_FEE: str = "fee"
_FIAT_DEPOSIT: str = "fiat_deposit"
_FIAT_WITHDRAWAL: str = "fiat_withdrawal"
_FILL_PRICE: str = "fill_price"
_FROM: str = "from"
_HASH: str = "hash"
_ID: str = "id"
_INTEREST: str = "interest"
_INFLATION_REWARD: str = "inflation_reward"
_NATIVE_AMOUNT: str = "native_amount"
_NETWORK: str = "network"
_OFF_BLOCKCHAIN: str = "off_blockchain"
_PRIME_WITHDRAWAL: str = "prime_withdrawal"
_PRO_DEPOSIT: str = "pro_deposit"
_PRO_WITHDRAWAL: str = "pro_withdrawal"
_RESOURCE: str = "resource"
_SELL: str = "sell"
_SEND: str = "send"
_STAKING_REWARD: str = "staking_reward"
_STATUS: str = "status"
_SUBTITLE: str = "subtitle"
_TITLE: str = "title"
_TO: str = "to"
_TRADE: str = "trade"
_TYPE: str = "type"
_UNIT_PRICE: str = "unit_price"
_UPDATED_AT: str = "updated_at"
_USER: str = "user"


class _ProcessAccountResult(NamedTuple):
    in_transactions: List[InTransaction]
    out_transactions: List[OutTransaction]
    intra_transactions: List[IntraTransaction]
    in_transaction_2_trade_id: Dict[InTransaction, str]
    trade_id_2_out_transaction: Dict[str, OutTransaction]
    out_transaction_2_trade_id: Dict[OutTransaction, str]
    trade_id_2_in_transaction: Dict[str, InTransaction]


class _InTransactionAndIndex(NamedTuple):
    in_transaction: InTransaction
    in_transaction_index: int


class _OutTransactionAndIndex(NamedTuple):
    out_transaction: OutTransaction
    out_transaction_index: int


class _SwapPair(NamedTuple):
    in_transaction: Optional[_InTransactionAndIndex]
    out_transaction: Optional[_OutTransactionAndIndex]


class _CoinbaseAuth(AuthBase):

    __API_VERSION: str = "2017-11-27"

    def __init__(self, api_key: str, api_secret: str) -> None:
        self.__api_key: str = api_key
        self.__api_secret: str = api_secret

    def __call__(self, request: PreparedRequest) -> PreparedRequest:
        timestamp: str = str(int(time.time()))
        message: str = f"{timestamp}{request.method}{request.path_url}{(request.body or '')}"  # type: ignore
        secret: str = self.__api_secret
        signature: str = hmac.new(secret.encode(), message.encode(), hashlib.sha256).hexdigest()
        request.headers.update(
            {
                "CB-VERSION": self.__API_VERSION,
                "CB-ACCESS-KEY": self.__api_key,
                "CB-ACCESS-SIGN": signature,
                "CB-ACCESS-TIMESTAMP": timestamp,
            }
        )
        return request


class InputPlugin(AbstractInputPlugin):

    __API_URL: str = "https://api.coinbase.com"
    __DEFAULT_THREAD_COUNT: int = 3
    __MAX_THREAD_COUNT: int = 4
    # Coinbase returns very low precision fiat data (only 2 decimal digits): if the value is less than 1c Coinbase
    # rounds it to zero, which causes various computation problems (spot_price, etc.). As a workaround, when this
    # condition is detected the plugin sets affected fields to UNKNOWN or None (depending on their nature), so that
    # they can be filled later by the transaction resolver and RP2.
    __MINIMUM_FIAT_PRECISION: RP2Decimal = RP2Decimal("0.01")
    __TIMEOUT: int = 30

    __COINBASE: str = "Coinbase"
    __COINBASE_PRO: str = "Coinbase Pro"

    def __init__(
        self,
        account_holder: str,
        api_key: str,
        api_secret: str,
        native_fiat: Optional[str] = None,
        thread_count: Optional[int] = None,
    ) -> None:

        super().__init__(account_holder=account_holder, native_fiat=native_fiat)
        self.__api_url: str = InputPlugin.__API_URL
        self.__auth: _CoinbaseAuth = _CoinbaseAuth(api_key, api_secret)
        self.__session: Session = requests.Session()
        self.__logger: logging.Logger = create_logger(f"{self.__COINBASE}/{self.account_holder}")
        self.__cache_key: str = f"{self.__COINBASE.lower()}-{account_holder}"
        self.__thread_count = thread_count if thread_count else self.__DEFAULT_THREAD_COUNT
        if self.__thread_count > self.__MAX_THREAD_COUNT:
            raise RP2RuntimeError(f"Thread count is {self.__thread_count}: it exceeds the maximum value of {self.__MAX_THREAD_COUNT}")

    def cache_key(self) -> Optional[str]:
        return self.__cache_key

    def load(self, country: AbstractCountry) -> List[AbstractTransaction]:
        result: List[AbstractTransaction] = []
        in_transaction_2_trade_id: Dict[InTransaction, str] = {}
        trade_id_2_out_transaction: Dict[str, OutTransaction] = {}
        out_transaction_2_trade_id: Dict[OutTransaction, str] = {}
        trade_id_2_in_transaction: Dict[str, InTransaction] = {}
        process_account_result_list: List[Optional[_ProcessAccountResult]]

        with ThreadPool(self.__thread_count) as pool:
            process_account_result_list = pool.map(self._process_account, self.__get_accounts())

        for process_account_result in process_account_result_list:
            if process_account_result is None:
                continue
            if process_account_result.in_transactions:
                result.extend(process_account_result.in_transactions)
            if process_account_result.out_transactions:
                result.extend(process_account_result.out_transactions)
            if process_account_result.intra_transactions:
                result.extend(process_account_result.intra_transactions)
            if process_account_result.in_transaction_2_trade_id:
                in_transaction_2_trade_id.update(process_account_result.in_transaction_2_trade_id)
            if process_account_result.trade_id_2_out_transaction:
                trade_id_2_out_transaction.update(process_account_result.trade_id_2_out_transaction)
            if process_account_result.out_transaction_2_trade_id:
                out_transaction_2_trade_id.update(process_account_result.out_transaction_2_trade_id)
            if process_account_result.trade_id_2_in_transaction:
                trade_id_2_in_transaction.update(process_account_result.trade_id_2_in_transaction)

        self._postprocess_swaps(
            transaction_list=result,
            in_transaction_2_trade_id=in_transaction_2_trade_id,
            trade_id_2_out_transaction=trade_id_2_out_transaction,
            out_transaction_2_trade_id=out_transaction_2_trade_id,
            trade_id_2_in_transaction=trade_id_2_in_transaction,
        )

        return result

    @staticmethod
    def _is_buy_side_of_swap(
        transaction: AbstractTransaction,
        in_transaction_2_trade_id: Dict[InTransaction, str],
        trade_id_2_out_transaction: Dict[str, OutTransaction],
    ) -> bool:
        return (
            isinstance(transaction, InTransaction)
            and transaction in in_transaction_2_trade_id
            and in_transaction_2_trade_id[transaction] in trade_id_2_out_transaction
        )

    @staticmethod
    def _is_sell_side_of_swap(
        transaction: AbstractTransaction,
        out_transaction_2_trade_id: Dict[OutTransaction, str],
        trade_id_2_in_transaction: Dict[str, InTransaction],
    ) -> bool:
        return (
            isinstance(transaction, OutTransaction)
            and transaction in out_transaction_2_trade_id
            and out_transaction_2_trade_id[transaction] in trade_id_2_in_transaction
        )

    # Crypto swap postprocessing (update fiat fields, notes, etc.)
    def _postprocess_swaps(
        self,
        transaction_list: List[AbstractTransaction],
        in_transaction_2_trade_id: Dict[InTransaction, str],
        trade_id_2_out_transaction: Dict[str, OutTransaction],
        out_transaction_2_trade_id: Dict[OutTransaction, str],
        trade_id_2_in_transaction: Dict[str, InTransaction],
    ) -> None:
        trade_id_2_swap_pair: Dict[str, _SwapPair] = {}
        in_transaction: InTransaction
        out_transaction: OutTransaction
        trade_id: str

        # Collect both sides of the swap
        for index, transaction in enumerate(transaction_list):
            swap_pair: _SwapPair
            if self._is_buy_side_of_swap(transaction, in_transaction_2_trade_id, trade_id_2_out_transaction):
                in_transaction = cast(InTransaction, transaction)
                trade_id = in_transaction_2_trade_id[in_transaction]
                out_transaction = trade_id_2_out_transaction[trade_id]
                if in_transaction.asset == out_transaction.asset:
                    raise RP2RuntimeError(
                        f"Internal error: detected a crypto swap with same asset ({in_transaction.asset}): {in_transaction} // {out_transaction}"
                    )
                in_transaction_and_index: _InTransactionAndIndex = _InTransactionAndIndex(
                    in_transaction=in_transaction,
                    in_transaction_index=index,
                )
                if trade_id in trade_id_2_swap_pair:
                    swap_pair = trade_id_2_swap_pair[trade_id]
                    swap_pair = _SwapPair(
                        in_transaction=in_transaction_and_index,
                        out_transaction=swap_pair.out_transaction,
                    )
                else:
                    swap_pair = _SwapPair(
                        in_transaction=in_transaction_and_index,
                        out_transaction=None,
                    )
            elif self._is_sell_side_of_swap(transaction, out_transaction_2_trade_id, trade_id_2_in_transaction):
                out_transaction = cast(OutTransaction, transaction)
                trade_id = out_transaction_2_trade_id[out_transaction]
                in_transaction = trade_id_2_in_transaction[trade_id]
                if in_transaction.asset == out_transaction.asset:
                    raise RP2RuntimeError(
                        f"Internal error: detected a crypto swap with same asset ({in_transaction.asset}): {in_transaction} // {out_transaction}"
                    )
                out_transaction_and_index: _OutTransactionAndIndex = _OutTransactionAndIndex(
                    out_transaction=out_transaction,
                    out_transaction_index=index,
                )
                if trade_id in trade_id_2_swap_pair:
                    swap_pair = trade_id_2_swap_pair[trade_id]
                    swap_pair = _SwapPair(
                        in_transaction=swap_pair.in_transaction,
                        out_transaction=out_transaction_and_index,
                    )
                else:
                    swap_pair = _SwapPair(
                        in_transaction=None,
                        out_transaction=out_transaction_and_index,
                    )
            else:
                # Not part of a swap
                continue

            trade_id_2_swap_pair[trade_id] = swap_pair

        # Process swaps
        for trade_id, swap_pair in trade_id_2_swap_pair.items():
            if swap_pair.in_transaction is None or swap_pair.out_transaction is None:
                raise RP2RuntimeError(f"Internal error: unmatched swap pair: {swap_pair}")
            in_transaction = swap_pair.in_transaction.in_transaction
            out_transaction = swap_pair.out_transaction.out_transaction

            # Ensure fees are not yet computed
            if in_transaction.crypto_fee is not None or in_transaction.fiat_fee is None or RP2Decimal(in_transaction.fiat_fee) != ZERO:
                raise RP2RuntimeError(f"Internal error: in-transaction crypto_fee is not None or fiat_fee != 0: {in_transaction}")
            if (
                out_transaction.crypto_fee is None
                or RP2Decimal(out_transaction.crypto_fee) != ZERO
                or out_transaction.fiat_fee is None
                or RP2Decimal(out_transaction.fiat_fee) != ZERO
            ):
                raise RP2RuntimeError(f"Internal error: out-transaction crypto_fee != 0 or fiat_fee != 0: {out_transaction}")

            fiat_out_no_fee: RP2Decimal
            # The fiat_fee is paid in the InTransaction, unless the InTransaction is fiat (which is ignored in RP2):
            # in this case the fiat_fee is paid in the OutTransaction
            if not self.is_native_fiat(in_transaction.asset):
                if not in_transaction.fiat_in_no_fee or not in_transaction.fiat_in_with_fee or not out_transaction.fiat_out_no_fee:
                    raise RP2RuntimeError(f"Internal error: swap transactions have incomplete fiat data: {in_transaction}//{out_transaction}")
                fiat_in_no_fee: RP2Decimal = RP2Decimal(in_transaction.fiat_in_no_fee)
                fiat_in_with_fee: RP2Decimal = RP2Decimal(in_transaction.fiat_in_with_fee)
                fiat_out_no_fee = RP2Decimal(out_transaction.fiat_out_no_fee)
                fiat_in_fee: RP2Decimal = fiat_out_no_fee - fiat_in_no_fee
                if fiat_in_with_fee > fiat_out_no_fee:
                    # Sometimes Coinbase produces conversions with fiat_in_with_fee > fiat_out_no_fee
                    # (e.g. see https://github.com/eprbell/dali-rp2/issues/34). This is clearly incorrect
                    # but the Coinbase API sometimes provides imprecise fiat data (see also
                    # https://github.com/eprbell/dali-rp2/issues/20 for more on this problem).
                    self.__logger.warning(
                        "Coinbase returned imprecise data: fiat_in_with_fee > fiat_out_no_fee: artificially setting fiat fee to 0: %s // %s",
                        str(in_transaction),
                        str(out_transaction),
                    )
                    fiat_in_fee = ZERO  # Set fiat_fee to ZERO, otherwise it would be negative

                # crypto_fee is always None (see swap logic in _process_fill()). Update notes and fiat_fee-related fields
                transaction_list[swap_pair.in_transaction.in_transaction_index] = InTransaction(
                    plugin=self.__COINBASE,
                    unique_id=in_transaction.unique_id,
                    raw_data=in_transaction.raw_data,
                    timestamp=in_transaction.timestamp,
                    asset=in_transaction.asset,
                    exchange=in_transaction.exchange,
                    holder=in_transaction.holder,
                    transaction_type=in_transaction.transaction_type,
                    spot_price=in_transaction.spot_price,
                    crypto_in=in_transaction.crypto_in,
                    crypto_fee=in_transaction.crypto_fee,
                    fiat_in_no_fee=in_transaction.fiat_in_no_fee,
                    fiat_in_with_fee=str(fiat_out_no_fee),
                    fiat_fee=str(fiat_in_fee),
                    notes=(
                        f"{in_transaction.notes + '; ' if in_transaction.notes else ''} Buy side of conversion from "
                        f"{out_transaction.crypto_out_with_fee} {out_transaction.asset} -> "
                        f"{in_transaction.crypto_in} {in_transaction.asset} "
                        f"({out_transaction.asset} out-transaction unique id: {out_transaction.unique_id})"
                    ),
                )
                # crypto_fee and fiat_fee are always 0 (see swap logic in _process_fill()). Update notes
                transaction_list[swap_pair.out_transaction.out_transaction_index] = OutTransaction(
                    plugin=self.__COINBASE,
                    unique_id=out_transaction.unique_id,
                    raw_data=out_transaction.raw_data,
                    timestamp=out_transaction.timestamp,
                    asset=out_transaction.asset,
                    exchange=out_transaction.exchange,
                    holder=out_transaction.holder,
                    transaction_type=out_transaction.transaction_type,
                    spot_price=out_transaction.spot_price,
                    crypto_out_no_fee=out_transaction.crypto_out_no_fee,
                    crypto_fee=out_transaction.crypto_fee,
                    crypto_out_with_fee=out_transaction.crypto_out_with_fee,
                    fiat_out_no_fee=out_transaction.fiat_out_no_fee,
                    fiat_fee=out_transaction.fiat_fee,
                    notes=(
                        f"{out_transaction.notes + '; ' if out_transaction.notes else ''} Sell side of conversion from "
                        f"{out_transaction.crypto_out_with_fee} {out_transaction.asset} -> "
                        f"{in_transaction.crypto_in} {in_transaction.asset} "
                        f"({in_transaction.asset} in-transaction unique id: {in_transaction.unique_id})"
                    ),
                )
            else:
                # InTransaction is fiat, so it is ignored in RP2: however any fiat_fee must be applied to the OutTransaction to avoid losing it
                if not in_transaction.fiat_fee or not out_transaction.fiat_out_no_fee:
                    raise RP2RuntimeError(f"Internal error: swap transactions have incomplete fiat data: {in_transaction} // {out_transaction}")
                fiat_out_no_fee = RP2Decimal(out_transaction.fiat_out_no_fee)
                fiat_fee: RP2Decimal = RP2Decimal(in_transaction.fiat_fee) if in_transaction.fiat_fee else ZERO
                transaction_list[swap_pair.out_transaction.out_transaction_index] = OutTransaction(
                    plugin=self.__COINBASE,
                    unique_id=out_transaction.unique_id,
                    raw_data=out_transaction.raw_data,
                    timestamp=out_transaction.timestamp,
                    asset=out_transaction.asset,
                    exchange=out_transaction.exchange,
                    holder=out_transaction.holder,
                    transaction_type=out_transaction.transaction_type,
                    spot_price=out_transaction.spot_price,
                    crypto_out_no_fee=out_transaction.crypto_out_no_fee,
                    crypto_fee=out_transaction.crypto_fee,
                    crypto_out_with_fee=out_transaction.crypto_out_with_fee,
                    fiat_out_no_fee=str(fiat_out_no_fee - fiat_fee),
                    fiat_fee=str(fiat_fee),
                    notes=(
                        f"{out_transaction.notes + '; ' if out_transaction.notes else ''} Sell side of conversion from "
                        f"{out_transaction.crypto_out_with_fee} {out_transaction.asset} -> "
                        f"{in_transaction.crypto_in} {in_transaction.asset} "
                        f"({in_transaction.asset} in-transaction unique id: {in_transaction.unique_id})"
                    ),
                )

    def _is_credit_card_spend(self, transaction: Any) -> bool:

        return (  # type: ignore
            transaction[_TYPE] is None
            and _TO in transaction
            and _EMAIL in transaction[_TO]
            and transaction[_TO][_EMAIL] == "treasury+coinbase-card@coinbase.com"
        ) or (
            transaction[_TYPE] == _CARDSPEND
        )

    def _process_account(self, account: Dict[str, Any]) -> Optional[_ProcessAccountResult]:
        currency: str = account[_CURRENCY][_CODE]
        account_id: str = account[_ID]
        in_transaction_list: List[InTransaction] = []
        out_transaction_list: List[OutTransaction] = []
        intra_transaction_list: List[IntraTransaction] = []
        in_transaction_2_trade_id: Dict[InTransaction, str] = {}
        trade_id_2_out_transaction: Dict[str, OutTransaction] = {}
        out_transaction_2_trade_id: Dict[OutTransaction, str] = {}
        trade_id_2_in_transaction: Dict[str, InTransaction] = {}

        self.__logger.debug("Account: %s", json.dumps(account))

        if account[_CREATED_AT] == account[_UPDATED_AT] and RP2Decimal(account[_BALANCE][_AMOUNT]) == ZERO:
            # skip account without activity to avoid unnecessary API calls
            return None

        id_2_buy: Dict[str, Any] = {}
        for buy in self.__get_buys(account_id):
            id_2_buy[buy[_ID]] = buy

        id_2_sell: Dict[str, Any] = {}
        for sell in self.__get_sells(account_id):
            id_2_sell[sell[_ID]] = sell

        for transaction in self.__get_transactions(account_id):
            raw_data: str = json.dumps(transaction)
            self.__logger.debug("Transaction: %s", raw_data)
            transaction_type: str = transaction[_TYPE]
            if transaction_type in {_PRIME_WITHDRAWAL, _PRO_DEPOSIT, _PRO_WITHDRAWAL, _EXCHANGE_DEPOSIT, _EXCHANGE_WITHDRAWAL, _SEND}:
                self._process_transfer(transaction, currency, in_transaction_list, out_transaction_list, intra_transaction_list)
            elif transaction_type in {_BUY, _SELL, _TRADE}:
                self._process_fill(
                    transaction=transaction,
                    currency=currency,
                    in_transaction_list=in_transaction_list,
                    out_transaction_list=out_transaction_list,
                    in_transaction_2_trade_id=in_transaction_2_trade_id,
                    trade_id_2_out_transaction=trade_id_2_out_transaction,
                    out_transaction_2_trade_id=out_transaction_2_trade_id,
                    trade_id_2_in_transaction=trade_id_2_in_transaction,
                    id_2_buy=id_2_buy,
                    id_2_sell=id_2_sell,
                )
            elif transaction_type in {_ADVANCED_TRADE_FILL}:
                self._process_advanced_trade_fill(
                    transaction=transaction,
                    currency=currency,
                    in_transaction_list=in_transaction_list,
                    out_transaction_list=out_transaction_list,
                )
            elif transaction_type in {_INTEREST}:
                self._process_gain(transaction, currency, Keyword.INTEREST, in_transaction_list)
            elif transaction_type in {_STAKING_REWARD}:
                self._process_gain(transaction, currency, Keyword.STAKING, in_transaction_list)
            elif transaction_type in {_INFLATION_REWARD}:
                self._process_gain(transaction, currency, Keyword.INCOME, in_transaction_list)
            elif transaction_type in {_FIAT_DEPOSIT, _CARDBUYBACK}:
                # _CARDBUYBACK is a credit card refund
                self._process_fiat_deposit(transaction, currency, in_transaction_list)
            elif transaction_type in {_FIAT_WITHDRAWAL}:
                self._process_fiat_withdrawal(transaction, currency, out_transaction_list)
            elif self._is_credit_card_spend(transaction):
                self._process_fiat_withdrawal(transaction, currency, out_transaction_list, "Coinbase card spend")
            else:
                self.__logger.error("Unsupported transaction type (skipping): %s. Please open an issue at %s", raw_data, self.ISSUES_URL)

        return _ProcessAccountResult(
            in_transactions=in_transaction_list,
            out_transactions=out_transaction_list,
            intra_transactions=intra_transaction_list,
            in_transaction_2_trade_id=in_transaction_2_trade_id,
            trade_id_2_out_transaction=trade_id_2_out_transaction,
            out_transaction_2_trade_id=out_transaction_2_trade_id,
            trade_id_2_in_transaction=trade_id_2_in_transaction,
        )

    def _process_transfer(
        self,
        transaction: Any,
        currency: str,
        in_transaction_list: List[InTransaction],
        out_transaction_list: List[OutTransaction],
        intra_transaction_list: List[IntraTransaction],
    ) -> None:
        # Ensure the amount is positive
        amount: RP2Decimal = RP2Decimal(transaction[_AMOUNT][_AMOUNT])
        native_amount: RP2Decimal = RP2Decimal(transaction[_NATIVE_AMOUNT][_AMOUNT])
        transaction_type: str = transaction[_TYPE]
        raw_data: str = json.dumps(transaction)

        if transaction_type in {_PRIME_WITHDRAWAL, _PRO_WITHDRAWAL, _EXCHANGE_WITHDRAWAL}:
            intra_transaction_list.append(
                IntraTransaction(
                    plugin=self.__COINBASE,
                    unique_id=transaction[_ID],
                    raw_data=raw_data,
                    timestamp=transaction[_CREATED_AT],
                    asset=currency,
                    from_exchange=self.__COINBASE_PRO,
                    from_holder=self.account_holder,
                    to_exchange=self.__COINBASE,
                    to_holder=self.account_holder,
                    spot_price=None,
                    crypto_sent=str(amount),
                    crypto_received=str(amount),
                )
            )
        elif transaction_type in [_PRO_DEPOSIT, _EXCHANGE_DEPOSIT]:
            intra_transaction_list.append(
                IntraTransaction(
                    plugin=self.__COINBASE,
                    unique_id=transaction[_ID],
                    raw_data=raw_data,
                    timestamp=transaction[_CREATED_AT],
                    asset=currency,
                    from_exchange=self.__COINBASE,
                    from_holder=self.account_holder,
                    to_exchange=self.__COINBASE_PRO,
                    to_holder=self.account_holder,
                    spot_price=None,
                    crypto_sent=str(-amount),
                    crypto_received=str(-amount),
                )
            )
        elif transaction_type == _SEND:
            transaction_network = transaction[_NETWORK]
            crypto_hash: str = transaction_network[_HASH] if _HASH in transaction_network else Keyword.UNKNOWN.value
            if amount < ZERO:
                if (
                    _TO in transaction
                    and transaction[_TO][_RESOURCE] == _USER
                    and transaction_network[_STATUS] == _OFF_BLOCKCHAIN
                    and _SUBTITLE in transaction[_DETAILS]
                    and _EMAIL in transaction[_TO]
                ):
                    # Outgoing gift to another Coinbase user
                    out_transaction_list.append(
                        OutTransaction(
                            plugin=self.__COINBASE,
                            unique_id=transaction[_ID],
                            raw_data=raw_data,
                            timestamp=transaction[_CREATED_AT],
                            asset=currency,
                            exchange=self.__COINBASE,
                            holder=self.account_holder,
                            transaction_type="Gift",
                            spot_price=str(native_amount / amount),
                            crypto_out_no_fee=str(-amount),
                            crypto_fee="0",
                            crypto_out_with_fee=str(-amount),
                            fiat_out_no_fee=str(-native_amount),
                            fiat_fee="0",
                            notes=f"To: {transaction[_TO][_EMAIL]}",
                        )
                    )
                elif _FROM in transaction and transaction[_DETAILS][_SUBTITLE].startswith("From Coinbase"):
                    # Coinbase Earn reversal transactions (due to credit card refunds typically): this is conservatively treated as a sale.
                    out_transaction_list.append(
                        OutTransaction(
                            plugin=self.__COINBASE,
                            unique_id=transaction[_ID],
                            raw_data=raw_data,
                            timestamp=transaction[_CREATED_AT],
                            asset=currency,
                            exchange=self.__COINBASE,
                            holder=self.account_holder,
                            transaction_type=Keyword.SELL.name,
                            spot_price=str(native_amount / amount),
                            crypto_out_no_fee=str(-amount),
                            crypto_fee="0",
                            crypto_out_with_fee=str(-amount),
                            fiat_out_no_fee=str(-native_amount),
                            fiat_fee="0",
                            notes="Coinbase EARN reversal",
                        )
                    )
                else:
                    intra_transaction_list.append(
                        IntraTransaction(
                            plugin=self.__COINBASE,
                            unique_id=crypto_hash,
                            raw_data=raw_data,
                            timestamp=transaction[_CREATED_AT],
                            asset=currency,
                            from_exchange=self.__COINBASE,
                            from_holder=self.account_holder,
                            to_exchange=Keyword.UNKNOWN.value,
                            to_holder=Keyword.UNKNOWN.value,
                            spot_price=str(native_amount / amount),
                            crypto_sent=str(-amount),
                            crypto_received=Keyword.UNKNOWN.value,
                        )
                    )
            else:
                if (
                    _FROM in transaction
                    and transaction[_FROM][_RESOURCE] == _USER
                    and transaction_network[_STATUS] == _OFF_BLOCKCHAIN
                    and _SUBTITLE in transaction[_DETAILS]
                ):
                    if _EMAIL in transaction[_FROM]:
                        # Incoming money from another Coinbase user. Marking it as income conservatively, but it could be
                        # a gift or other type: if so the user needs to explicitly recast it with a transaction hint
                        self._process_gain(transaction, currency, Keyword.INCOME, in_transaction_list, f"From: {transaction[_FROM][_EMAIL]}")
                    elif transaction[_DETAILS][_SUBTITLE].startswith("From Coinbase"):
                        # Coinbase Earn transactions
                        self._process_gain(transaction, currency, Keyword.INCOME, in_transaction_list, "Coinbase EARN")
                else:
                    intra_transaction_list.append(
                        IntraTransaction(
                            plugin=self.__COINBASE,
                            unique_id=crypto_hash,
                            raw_data=raw_data,
                            timestamp=transaction[_CREATED_AT],
                            asset=currency,
                            from_exchange=Keyword.UNKNOWN.value,
                            from_holder=Keyword.UNKNOWN.value,
                            to_exchange=self.__COINBASE,
                            to_holder=self.account_holder,
                            spot_price=str(native_amount / amount),
                            crypto_sent=Keyword.UNKNOWN.value,
                            crypto_received=str(amount),
                        )
                    )

    def _process_fill(
        self,
        transaction: Any,
        currency: str,
        in_transaction_list: List[InTransaction],
        out_transaction_list: List[OutTransaction],
        in_transaction_2_trade_id: Dict[InTransaction, str],
        trade_id_2_out_transaction: Dict[str, OutTransaction],
        out_transaction_2_trade_id: Dict[OutTransaction, str],
        trade_id_2_in_transaction: Dict[str, InTransaction],
        id_2_buy: Dict[str, Any],
        id_2_sell: Dict[str, Any],
    ) -> None:
        transaction_type: str = transaction[_TYPE]
        native_amount: RP2Decimal = RP2Decimal(transaction[_NATIVE_AMOUNT][_AMOUNT])
        crypto_amount: RP2Decimal = RP2Decimal(transaction[_AMOUNT][_AMOUNT])
        fiat_fee: RP2Decimal = ZERO
        spot_price: RP2Decimal
        spot_price_string: str

        if not self.is_native_fiat(transaction[_NATIVE_AMOUNT][_CURRENCY]):
            raise RP2RuntimeError(f"Internal error: native amount is not denominated in {self.native_fiat} {json.dumps(transaction)}")

        raw_data: str = json.dumps(transaction)
        if not transaction_type in {_BUY, _SELL, _TRADE}:
            self.__logger.error("Unsupported transaction type for fill (skipping): %s. Please open an issue at %s", raw_data, self.ISSUES_URL)
            return

        if native_amount >= ZERO:
            if transaction_type == _BUY:
                buy: Dict[str, Any] = id_2_buy[transaction[transaction_type][_ID]]
                raw_data = f"{raw_data}//{json.dumps(buy)}"
                fiat_fee = RP2Decimal(buy[_FEE][_AMOUNT])
                spot_price = RP2Decimal(buy[_UNIT_PRICE][_AMOUNT])
                spot_price_string = str(spot_price)
                self.__logger.debug("Buy: %s", json.dumps(buy))
            else:
                # swap in transaction
                spot_price_string = Keyword.UNKNOWN.value
                if native_amount >= self.__MINIMUM_FIAT_PRECISION:
                    spot_price_string = str((native_amount - fiat_fee) / crypto_amount)

            fiat_in_no_fee: Optional[str] = None
            fiat_in_with_fee: Optional[str] = None
            if native_amount >= self.__MINIMUM_FIAT_PRECISION:
                fiat_in_no_fee = str(native_amount - fiat_fee)
                fiat_in_with_fee = str(native_amount)

            in_transaction: InTransaction = InTransaction(
                plugin=self.__COINBASE,
                unique_id=transaction[_ID],
                raw_data=raw_data,
                timestamp=transaction[_CREATED_AT],
                asset=currency,
                exchange=self.__COINBASE,
                holder=self.account_holder,
                transaction_type=Keyword.BUY.name,
                spot_price=spot_price_string,
                crypto_in=str(crypto_amount),
                crypto_fee=None,
                fiat_in_no_fee=fiat_in_no_fee,
                fiat_in_with_fee=fiat_in_with_fee,
                fiat_fee=str(fiat_fee),
                notes=None,
            )
            in_transaction_list.append(in_transaction)
            if transaction_type == _TRADE:
                in_transaction_2_trade_id[in_transaction] = transaction[_TRADE][_ID]
                trade_id_2_in_transaction[transaction[_TRADE][_ID]] = in_transaction

        elif native_amount < ZERO:
            if transaction_type == _SELL:
                sell = id_2_sell[transaction[transaction_type][_ID]]
                raw_data = f"{raw_data}//{json.dumps(sell)}"
                fiat_fee = RP2Decimal(sell[_FEE][_AMOUNT])
                spot_price = RP2Decimal(sell[_UNIT_PRICE][_AMOUNT])
                self.__logger.debug("Sell: %s", json.dumps(sell))
            else:
                # swap out transaction
                spot_price = native_amount / crypto_amount

            out_transaction: OutTransaction = OutTransaction(
                plugin=self.__COINBASE,
                unique_id=transaction[_ID],
                raw_data=raw_data,
                timestamp=transaction[_CREATED_AT],
                asset=currency,
                exchange=self.__COINBASE,
                holder=self.account_holder,
                transaction_type=Keyword.SELL.name,
                spot_price=str(spot_price),
                crypto_out_no_fee=str(-crypto_amount - fiat_fee / spot_price),
                crypto_fee=str(fiat_fee / spot_price),
                crypto_out_with_fee=str(-crypto_amount),
                fiat_out_no_fee=str(-native_amount),
                fiat_fee=str(fiat_fee),
                notes=None,
            )
            out_transaction_list.append(out_transaction)
            if transaction_type == _TRADE:
                out_transaction_2_trade_id[out_transaction] = transaction[_TRADE][_ID]
                trade_id_2_out_transaction[transaction[_TRADE][_ID]] = out_transaction

    def _process_advanced_trade_fill(
        self,
        transaction: Any,
        currency: str,
        in_transaction_list: List[InTransaction],
        out_transaction_list: List[OutTransaction],
    ) -> None:
        transaction_type: str = transaction[_TYPE]
        native_amount: RP2Decimal = RP2Decimal(transaction[_NATIVE_AMOUNT][_AMOUNT])
        crypto_amount: RP2Decimal = RP2Decimal(transaction[_AMOUNT][_AMOUNT])
        fiat_fee: RP2Decimal = ZERO
        spot_price: RP2Decimal

        if not self.is_native_fiat(transaction[_NATIVE_AMOUNT][_CURRENCY]):
            raise RP2RuntimeError(f"Internal error: native amount is not denominated in {self.native_fiat} {json.dumps(transaction)}")

        raw_data: str = json.dumps(transaction)
        if not transaction_type in {_ADVANCED_TRADE_FILL}:
            self.__logger.error("Unsupported transaction type for fill (skipping): %s. Please open an issue at %s", raw_data, self.ISSUES_URL)
            return

        if native_amount >= ZERO:
            raw_data = f"{raw_data}"
            fiat_fee = RP2Decimal(transaction[transaction_type][_COMMISSION])
            spot_price = RP2Decimal(transaction[transaction_type][_FILL_PRICE])

            fiat_in_no_fee: Optional[str] = None
            fiat_in_with_fee: Optional[str] = None
            if native_amount >= self.__MINIMUM_FIAT_PRECISION:
                fiat_in_no_fee = str(native_amount - fiat_fee)
                fiat_in_with_fee = str(native_amount)

            in_transaction: InTransaction = InTransaction(
                plugin=self.__COINBASE,
                unique_id=transaction[_ID],
                raw_data=raw_data,
                timestamp=transaction[_CREATED_AT],
                asset=currency,
                exchange=self.__COINBASE,
                holder=self.account_holder,
                transaction_type=Keyword.BUY.name,
                spot_price=str(spot_price),
                crypto_in=str(crypto_amount),
                crypto_fee=None,
                fiat_in_no_fee=fiat_in_no_fee,
                fiat_in_with_fee=fiat_in_with_fee,
                fiat_fee=str(fiat_fee),
                notes=None,
            )
            in_transaction_list.append(in_transaction)

        elif native_amount < ZERO:
            raw_data = f"{raw_data}"
            fiat_fee = RP2Decimal(transaction[transaction_type][_COMMISSION])
            spot_price = RP2Decimal(transaction[transaction_type][_FILL_PRICE])

            out_transaction: OutTransaction = OutTransaction(
                plugin=self.__COINBASE,
                unique_id=transaction[_ID],
                raw_data=raw_data,
                timestamp=transaction[_CREATED_AT],
                asset=currency,
                exchange=self.__COINBASE,
                holder=self.account_holder,
                transaction_type=Keyword.SELL.name,
                spot_price=str(spot_price),
                crypto_out_no_fee=str(-crypto_amount - fiat_fee / spot_price),
                crypto_fee=str(fiat_fee / spot_price),
                crypto_out_with_fee=str(-crypto_amount),
                fiat_out_no_fee=str(-native_amount),
                fiat_fee=str(fiat_fee),
                notes=None,
            )
            out_transaction_list.append(out_transaction)

    def _process_gain(
        self, transaction: Any, currency: str, transaction_type: Keyword, in_transaction_list: List[InTransaction], notes: Optional[str] = None
    ) -> None:
        amount: RP2Decimal = RP2Decimal(transaction[_AMOUNT][_AMOUNT])
        native_amount: RP2Decimal = RP2Decimal(transaction[_NATIVE_AMOUNT][_AMOUNT])
        notes = f"{notes + '; ' if notes else ''}{transaction[_DETAILS][_TITLE]}"
        spot_price: str = Keyword.UNKNOWN.value
        fiat_in_no_fee: Optional[str] = None
        fiat_in_with_fee: Optional[str] = None
        if native_amount >= self.__MINIMUM_FIAT_PRECISION:
            spot_price = str(native_amount / amount)
            fiat_in_no_fee = str(native_amount)
            fiat_in_with_fee = str(native_amount)
        in_transaction_list.append(
            InTransaction(
                plugin=self.__COINBASE,
                unique_id=transaction[_ID],
                raw_data=json.dumps(transaction),
                timestamp=transaction[_CREATED_AT],
                asset=currency,
                exchange=self.__COINBASE,
                holder=self.account_holder,
                transaction_type=transaction_type.value.capitalize(),
                spot_price=spot_price,
                crypto_in=str(amount),
                crypto_fee=None,
                fiat_in_no_fee=fiat_in_no_fee,
                fiat_in_with_fee=fiat_in_with_fee,
                fiat_fee="0",
                notes=notes,
            )
        )

    def _process_fiat_deposit(self, transaction: Any, currency: str, in_transaction_list: List[InTransaction], notes: Optional[str] = None) -> None:
        amount: RP2Decimal = RP2Decimal(transaction[_AMOUNT][_AMOUNT])
        details_title = transaction[_DETAILS][_TITLE]
        details_subtitle = transaction[_DETAILS][_SUBTITLE]
        notes = f"{notes + '; ' if notes else ''}{details_title + '; ' if details_title else ''}{details_subtitle if details_subtitle else ''}"
        in_transaction_list.append(
            InTransaction(
                plugin=self.__COINBASE,
                unique_id=transaction[_ID],
                raw_data=json.dumps(transaction),
                timestamp=transaction[_CREATED_AT],
                asset=currency,
                exchange=self.__COINBASE,
                holder=self.account_holder,
                transaction_type=Keyword.BUY.value,
                spot_price="1",
                crypto_in=str(amount),
                crypto_fee="0",
                fiat_in_no_fee=None,
                fiat_in_with_fee=None,
                fiat_fee=None,
                notes=notes,
            )
        )

    def _process_fiat_withdrawal(self, transaction: Any, currency: str, out_transaction_list: List[OutTransaction], notes: Optional[str] = None) -> None:
        amount: RP2Decimal = RP2Decimal(transaction[_AMOUNT][_AMOUNT])
        details_title = transaction[_DETAILS][_TITLE]
        details_subtitle = transaction[_DETAILS][_SUBTITLE]
        notes = f"{notes + '; ' if notes else ''}{details_title + '; ' if details_title else ''}{details_subtitle if details_subtitle else ''}"
        out_transaction_list.append(
            OutTransaction(
                plugin=self.__COINBASE,
                unique_id=transaction[_ID],
                raw_data=json.dumps(transaction),
                timestamp=transaction[_CREATED_AT],
                asset=currency,
                exchange=self.__COINBASE,
                holder=self.account_holder,
                transaction_type=Keyword.SELL.value,
                spot_price="1",
                crypto_out_no_fee=str(-amount),
                crypto_fee="0",
                crypto_out_with_fee=None,
                fiat_out_no_fee=None,
                fiat_fee=None,
                notes=notes,
            )
        )

    def __get_accounts(self) -> Any:
        return self.__send_request_with_pagination("/v2/accounts")

    def __get_transactions(self, account_id: str) -> Any:
        return self.__send_request_with_pagination(f"/v2/accounts/{account_id}/transactions")

    def __get_buys(self, account_id: str) -> Any:
        return self.__send_request_with_pagination(f"/v2/accounts/{account_id}/buys")

    def __get_sells(self, account_id: str) -> Any:
        return self.__send_request_with_pagination(f"/v2/accounts/{account_id}/sells")

    def __send_request(self, method: str, endpoint: str, params: Any = None, data: Any = None) -> Any:  # pylint: disable=unused-private-member
        full_url: str = f"{self.__api_url}{endpoint}"
        response: Response = self.__session.request(method, full_url, params=params, data=data, auth=self.__auth, timeout=self.__TIMEOUT)
        self._validate_response(response, method, endpoint)
        return response.json()

    # Documented at: https://docs.cloud.coinbase.com/exchange/docs/pagination
    def __send_request_with_pagination(self, endpoint: str, params: Optional[Dict[str, Any]] = None) -> Any:
        if params is None:
            params = {}
        current_url: str = f"{self.__api_url}{endpoint}"
        while True:
            response: Response = self.__session.get(current_url, params=params, auth=self.__auth, timeout=self.__TIMEOUT)
            self._validate_response(response, "get", endpoint)
            json_response: Any = response.json()
            yield from json_response["data"]
            if "pagination" not in json_response or "next_uri" not in json_response["pagination"] or not json_response["pagination"]["next_uri"]:
                break
            current_url = f"{self.__api_url}{json_response['pagination']['next_uri']}"

    # Documented at: https://docs.cloud.coinbase.com/exchange/docs/requests
    def _validate_response(self, response: Response, method: str, endpoint: str) -> None:
        json_response: Any = response.json()
        message: str = ""
        if 200 <= response.status_code < 300:
            return
        if "message" in json_response:
            message = json_response["message"]
            self.__logger.error("Error %d: %s%s (%s): %s", response.status_code, self.__api_url, endpoint, method.upper(), json_response["message"])

        response.raise_for_status()

        # Defensive programming: we shouldn't reach here.
        self.__logger.debug("Reached past raise_for_status() call: %s", json_response["message"])
        raise RP2RuntimeError(message)



# Copyright 2022 Steve Davis
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from datetime import datetime, timedelta
from typing import NamedTuple, cast

from rp2.rp2_decimal import RP2Decimal
from rp2.rp2_error import RP2ValueError

from dali.configuration import Keyword


class HistoricalBar(NamedTuple):
    """A single bar of historical market data."""

    duration: timedelta
    timestamp: datetime
    open: RP2Decimal
    high: RP2Decimal
    low: RP2Decimal
    close: RP2Decimal
    volume: RP2Decimal

    def derive_transaction_price(self, transaction_timestamp: datetime, historical_price_type: str) -> RP2Decimal:
        """Derive a transaction price from a historical bar."""
        price: RP2Decimal = cast(RP2Decimal, None)
        if historical_price_type == Keyword.HISTORICAL_PRICE_OPEN.value:
            price = self.open
        elif historical_price_type == Keyword.HISTORICAL_PRICE_HIGH.value:
            price = self.high
        elif historical_price_type == Keyword.HISTORICAL_PRICE_LOW.value:
            price = self.low
        elif historical_price_type == Keyword.HISTORICAL_PRICE_CLOSE.value:
            price = self.close
        elif historical_price_type == Keyword.HISTORICAL_PRICE_NEAREST.value:
            bar_start_timestamp = self.timestamp
            bar_end_timestamp = self.timestamp + self.duration
            start_timedelta = abs(transaction_timestamp - bar_start_timestamp)
            end_timedelta = abs(transaction_timestamp - bar_end_timestamp)
            price = self.open if start_timedelta < end_timedelta else self.close
        else:
            raise RP2ValueError(f"Unrecognized historical_price_type '{historical_price_type}'")
        return price



# Copyright 2022 macanudo527
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# This plugin uses the sheet found under the "for-cointracker" because of its simpler format.
# Trades CSV Format: timestamp UTC, received quantity, received currency, sent quantity,
#    sent currency, fee amount, fee currency, tag

# The transfers CSV can be found under the "depositwithdrawal" tab
# transfer CSV format: timestamp UTC, transaction type, amount, coin (prefixed with the network used), network, txid, fee

import logging
from csv import reader
from typing import List, Optional

from rp2.abstract_country import AbstractCountry
from rp2.logger import create_logger

from dali.abstract_input_plugin import AbstractInputPlugin
from dali.abstract_transaction import AbstractTransaction
from dali.configuration import Keyword
from dali.in_transaction import InTransaction
from dali.intra_transaction import IntraTransaction
from dali.out_transaction import OutTransaction


class InputPlugin(AbstractInputPlugin):

    __PIONEX: str = "Pionex"
    __PIONEX_PLUGIN: str = "Pionex_CSV"

    __TIMESTAMP_INDEX: int = 0
    __RECEIVED_AMOUNT: int = 1
    __TRANSACTION_TYPE: int = 1
    __ASSET_RECEIVED: int = 2
    __AMOUNT_TRANSFERED: int = 2
    __SENT_AMOUNT: int = 3
    __ASSET_TRANSFERED: int = 3
    __ASSET_SENT: int = 4
    __CHAIN_USED: int = 4
    __FEE_AMOUNT: int = 5
    __TXN_ID: int = 5
    __FEE_ASSET: int = 6

    __DELIMITER: str = ","

    # Keywords
    __DEPOSIT: str = "DEPOSIT"
    __WITHDRAWAL: str = "WITHDRAWAL"

    def __init__(
        self,
        account_holder: str,
        trades_csv_file: Optional[str] = None,
        transfers_csv_file: Optional[str] = None,
        native_fiat: Optional[str] = None,
    ) -> None:

        super().__init__(account_holder=account_holder, native_fiat=native_fiat)
        self.__trades_csv_file: Optional[str] = trades_csv_file
        self.__transfers_csv_file: Optional[str] = transfers_csv_file
        self.__logger: logging.Logger = create_logger(f"{self.__PIONEX_PLUGIN}/{self.account_holder}")

    def load(self, country: AbstractCountry) -> List[AbstractTransaction]:
        result: List[AbstractTransaction] = []

        if self.__trades_csv_file:
            result += self.parse_trades_file(self.__trades_csv_file)

        if self.__transfers_csv_file:
            result += self.parse_transfers_file(self.__transfers_csv_file)

        return result

    def parse_trades_file(self, file_path: str) -> List[AbstractTransaction]:
        result: List[AbstractTransaction] = []

        with open(file_path, encoding="utf-8") as csv_file:
            lines = reader(csv_file)

            header = next(lines)
            self.__logger.debug("Header: %s", header)
            for line in lines:
                # If there is a blank sent/receive asset, this is a transfer, which we will process under transfers
                # Pionex sometimes creates 0 entries for some reason
                if line[self.__ASSET_SENT] == "" or line[self.__ASSET_RECEIVED] == "" or float(line[self.__RECEIVED_AMOUNT]) == 0:
                    continue

                raw_data: str = self.__DELIMITER.join(line)
                self.__logger.debug("Transaction: %s", raw_data)

                in_crypto_fee: str = "0"
                out_crypto_fee: str = "0"

                if line[self.__ASSET_RECEIVED] == line[self.__FEE_ASSET]:
                    in_crypto_fee = line[self.__FEE_AMOUNT]
                else:
                    out_crypto_fee = line[self.__FEE_AMOUNT]

                result.append(
                    InTransaction(
                        plugin=self.__PIONEX_PLUGIN,
                        unique_id=Keyword.UNKNOWN.value,
                        raw_data=raw_data,
                        timestamp=f"{line[self.__TIMESTAMP_INDEX]} -00:00",
                        asset=line[self.__ASSET_RECEIVED],
                        exchange=self.__PIONEX,
                        holder=self.account_holder,
                        transaction_type=Keyword.BUY.value,
                        spot_price=Keyword.UNKNOWN.value,
                        crypto_in=line[self.__RECEIVED_AMOUNT],
                        crypto_fee=in_crypto_fee,
                        notes=None,
                    )
                )

                result.append(
                    OutTransaction(
                        plugin=self.__PIONEX_PLUGIN,
                        unique_id=Keyword.UNKNOWN.value,
                        raw_data=raw_data,
                        timestamp=f"{line[self.__TIMESTAMP_INDEX]} -00:00",
                        asset=line[self.__ASSET_SENT],
                        exchange=self.__PIONEX,
                        holder=self.account_holder,
                        transaction_type=Keyword.SELL.value,
                        spot_price=Keyword.UNKNOWN.value,
                        crypto_out_no_fee=line[self.__SENT_AMOUNT],
                        crypto_out_with_fee=line[self.__SENT_AMOUNT],
                        crypto_fee=out_crypto_fee,
                        notes=None,
                    )
                )

            return result

    def parse_transfers_file(self, file_path: str) -> List[AbstractTransaction]:
        result: List[AbstractTransaction] = []

        with open(file_path, encoding="utf-8") as csv_file:
            lines = reader(csv_file)

            header = next(lines)
            self.__logger.debug("Header: %s", header)
            for line in lines:

                raw_data: str = self.__DELIMITER.join(line)
                self.__logger.debug("Transaction: %s", raw_data)

                asset: str = (
                    line[self.__ASSET_TRANSFERED][: -len(line[self.__CHAIN_USED])]
                    if (line[self.__ASSET_TRANSFERED].endswith(line[self.__CHAIN_USED]))
                    else (line[self.__ASSET_TRANSFERED])
                )

                if line[self.__TRANSACTION_TYPE] == self.__DEPOSIT:
                    result.append(
                        IntraTransaction(
                            plugin=self.__PIONEX_PLUGIN,
                            unique_id=line[self.__TXN_ID],
                            raw_data=raw_data,
                            timestamp=f"{line[self.__TIMESTAMP_INDEX]} -00:00",
                            asset=asset,
                            from_exchange=Keyword.UNKNOWN.value,
                            from_holder=Keyword.UNKNOWN.value,
                            to_exchange=self.__PIONEX,
                            to_holder=self.account_holder,
                            spot_price=Keyword.UNKNOWN.value,
                            crypto_sent=Keyword.UNKNOWN.value,
                            crypto_received=str(line[self.__AMOUNT_TRANSFERED]),
                        )
                    )
                elif line[self.__TRANSACTION_TYPE] == self.__WITHDRAWAL:
                    result.append(
                        IntraTransaction(
                            plugin=self.__PIONEX_PLUGIN,
                            unique_id=line[self.__TXN_ID],
                            raw_data=raw_data,
                            timestamp=f"{line[self.__TIMESTAMP_INDEX]} -00:00",
                            asset=asset,
                            from_exchange=self.__PIONEX,
                            from_holder=self.account_holder,
                            to_exchange=Keyword.UNKNOWN.value,
                            to_holder=Keyword.UNKNOWN.value,
                            spot_price=Keyword.UNKNOWN.value,
                            crypto_sent=str(line[self.__AMOUNT_TRANSFERED]),
                            crypto_received=Keyword.UNKNOWN.value,
                        )
                    )
                else:
                    self.__logger.error("Unrecognized Crypto transfer: %s", raw_data)

            return result



# Copyright 2022 macanudo527
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Kraken CSV format: (epoch) timestamp, open, high, low, close, volume, trades

import logging
from csv import reader, writer
from datetime import datetime, timedelta, timezone
from gzip import open as gopen
from io import BytesIO
from json import JSONDecodeError
from multiprocessing.pool import ThreadPool
from os import makedirs, path
from typing import Any, Dict, Generator, List, NamedTuple, Optional, Tuple, cast
from zipfile import ZipFile

import requests
from requests.models import Response
from requests.sessions import Session
from rp2.logger import create_logger
from rp2.rp2_decimal import RP2Decimal
from rp2.rp2_error import RP2RuntimeError

from dali.cache import load_from_cache, save_to_cache
from dali.historical_bar import HistoricalBar

# Google Drive parameters
_ACCESS_NOT_CONFIGURED: str = "accessNotConfigured"
_BAD_REQUEST: str = "badRequest"
_ERROR: str = "error"
_ERRORS: str = "errors"
_FILES: str = "files"

# The endpoint we will use to query Google Drive for the specific file we need
# We will also use this to request a file download.
_GOOGLE_APIS_URL: str = "https://www.googleapis.com/drive/v3/files"
_ID: str = "id"

# File ID for the folder which contains zipped OHLCVT files grouped by asset
# In Google Drive lingo, this is the 'parent' of the files we need.
# Files in this folder will be updated every quarter and thus will have new file IDs.
# However, the file ID for the folder or parent should remain the same.
_KRAKEN_FOLDER_ID: str = "1aoA6SKgPbS_p3pYStXUXFvmjqShJ2jv9"
_MESSAGE: str = "message"
_REASON: str = "reason"

# Google Drive URL Params
_ALT: str = "alt"
_API_KEY: str = "key"
_CONFIRM: str = "confirm"
_MEDIA: str = "media"
_QUERY: str = "q"

# Time periods
_MS_IN_SECOND: int = 1000
_SECONDS_IN_DAY: int = 86400
_CHUNK_SIZE: int = _SECONDS_IN_DAY * 30
_SECONDS_IN_MINUTE: int = 60

# Time in minutes
_MINUTE: str = "1"
_FIVE_MINUTE: str = "5"
_FIFTEEN_MINUTE: str = "15"
_ONE_HOUR: str = "60"
_TWELVE_HOUR: str = "720"
_ONE_DAY: str = "1440"
_TIME_GRANULARITY: List[str] = [_MINUTE, _FIVE_MINUTE, _FIFTEEN_MINUTE, _ONE_HOUR, _TWELVE_HOUR, _ONE_DAY]

# Chunking variables
_PAIR_START: str = "start"
_PAIR_MIDDLE: str = "middle"
_PAIR_END: str = "end"
_MAX_MULTIPLIER: int = 500


class _PairStartEnd(NamedTuple):
    end: int
    start: int


class Kraken:

    __KRAKEN_OHLCVT: str = "Kraken.com_CSVOHLCVT"

    __CACHE_DIRECTORY: str = ".dali_cache/kraken/"
    __CACHE_KEY: str = "Kraken-csv-download"

    __TIMEOUT: int = 30
    __THREAD_COUNT: int = 3

    __TIMESTAMP_INDEX: int = 0
    __OPEN: int = 1
    __HIGH: int = 2
    __LOW: int = 3
    __CLOSE: int = 4
    __VOLUME: int = 5
    __TRADES: int = 6

    __DELIMITER: str = ","

    def __init__(
        self,
        google_api_key: str,
    ) -> None:

        self.__google_api_key: str = google_api_key
        self.__logger: logging.Logger = create_logger(self.__KRAKEN_OHLCVT)
        self.__session: Session = requests.Session()
        self.__cached_pairs: Dict[str, _PairStartEnd] = {}
        self.__cache_loaded: bool = False

        if not path.exists(self.__CACHE_DIRECTORY):
            makedirs(self.__CACHE_DIRECTORY)

    def cache_key(self) -> str:
        return self.__CACHE_KEY

    def __load_cache(self) -> None:
        result = cast(Dict[str, _PairStartEnd], load_from_cache(self.cache_key()))
        self.__cached_pairs = result if result is not None else {}

    def __split_process(self, csv_file: str, chunk_size: int = _CHUNK_SIZE) -> Generator[Tuple[str, List[List[str]]], None, None]:
        chunk: List[List[str]] = []

        lines = reader(csv_file.splitlines())
        position = _PAIR_START
        next_timestamp: Optional[int] = None

        for line in lines:
            if next_timestamp is None:
                next_timestamp = ((int(line[self.__TIMESTAMP_INDEX]) + chunk_size) // chunk_size) * chunk_size

            if int(line[self.__TIMESTAMP_INDEX]) % chunk_size == 0 or int(line[self.__TIMESTAMP_INDEX]) > next_timestamp:
                yield position, chunk
                if position == _PAIR_START:
                    position = _PAIR_MIDDLE
                chunk = []
                next_timestamp += chunk_size
            chunk.append(line)
        if chunk:
            position = _PAIR_END
            yield position, chunk

    def _split_chunks_size_n(self, file_name: str, csv_file: str, chunk_size: int = _CHUNK_SIZE) -> None:

        pair, duration_in_minutes = file_name.strip(".csv").split("_", 1)
        chunk_size *= min(int(duration_in_minutes), _MAX_MULTIPLIER)
        file_timestamp: str
        pair_start: Optional[int] = None
        pair_end: int
        pair_duration: str = pair + duration_in_minutes

        for position, chunk in self.__split_process(csv_file, chunk_size):
            file_timestamp = str((int(chunk[0][self.__TIMESTAMP_INDEX])) // chunk_size * chunk_size)
            if position == _PAIR_END:
                pair_end = int(chunk[-1][self.__TIMESTAMP_INDEX])
                if pair_start is None:
                    pair_start = int(chunk[0][self.__TIMESTAMP_INDEX])
            elif position == _PAIR_START:
                pair_start = int(chunk[0][self.__TIMESTAMP_INDEX])

            chunk_filename: str = f'{pair}_{file_timestamp}_{duration_in_minutes}.{"csv.gz"}'
            chunk_filepath: str = path.join(self.__CACHE_DIRECTORY, chunk_filename)

            with gopen(chunk_filepath, "wt", encoding="utf-8", newline="") as chunk_file:
                csv_writer = writer(chunk_file)
                for row in chunk:
                    csv_writer.writerow(row)

        if pair_start:
            self.__cached_pairs[pair_duration] = _PairStartEnd(start=pair_start, end=pair_end)

    def _retrieve_cached_bar(self, base_asset: str, quote_asset: str, timestamp: int) -> Optional[HistoricalBar]:
        pair_name: str = base_asset + quote_asset

        retry_count: int = 0

        while retry_count < len(_TIME_GRANULARITY):
            if (
                timestamp < self.__cached_pairs[pair_name + _TIME_GRANULARITY[retry_count]].start
                or timestamp > self.__cached_pairs[pair_name + _TIME_GRANULARITY[retry_count]].end
            ):
                self.__logger.debug(
                    "Out of range - %s < %s or %s > %s",
                    timestamp,
                    self.__cached_pairs[pair_name + _TIME_GRANULARITY[retry_count]].start,
                    timestamp,
                    self.__cached_pairs[pair_name + _TIME_GRANULARITY[retry_count]].end,
                )
                retry_count += 1
                continue

            duration_chunk_size = _CHUNK_SIZE * min(int(_TIME_GRANULARITY[retry_count]), _MAX_MULTIPLIER)
            file_timestamp: int = (timestamp // duration_chunk_size) * duration_chunk_size

            # Floor the timestamp to find the price
            duration_timestamp: int = (timestamp // (int(_TIME_GRANULARITY[retry_count]) * _SECONDS_IN_MINUTE)) * (
                int(_TIME_GRANULARITY[retry_count]) * _SECONDS_IN_MINUTE
            )

            file_name: str = f"{base_asset + quote_asset}_{file_timestamp}_{_TIME_GRANULARITY[retry_count]}.csv.gz"
            file_path: str = path.join(self.__CACHE_DIRECTORY, file_name)
            self.__logger.debug("Retrieving %s -> %s at %s from %s stamped file.", base_asset, quote_asset, duration_timestamp, file_timestamp)
            with gopen(file_path, "rt") as file:
                rows = reader(file)
                for row in rows:
                    if int(row[self.__TIMESTAMP_INDEX]) == duration_timestamp:
                        return HistoricalBar(
                            duration=timedelta(minutes=int(_TIME_GRANULARITY[retry_count])),
                            timestamp=datetime.fromtimestamp(int(row[self.__TIMESTAMP_INDEX]), timezone.utc),
                            open=RP2Decimal(row[self.__OPEN]),
                            high=RP2Decimal(row[self.__HIGH]),
                            low=RP2Decimal(row[self.__LOW]),
                            close=RP2Decimal(row[self.__CLOSE]),
                            volume=RP2Decimal(row[self.__VOLUME]),
                        )

            retry_count += 1

        return None

    def find_historical_bar(self, base_asset: str, quote_asset: str, timestamp: datetime) -> Optional[HistoricalBar]:
        epoch_timestamp = int(timestamp.timestamp())
        self.__logger.debug("Retrieving bar for %s%s at %s", base_asset, quote_asset, epoch_timestamp)

        if not self.__cache_loaded:
            self.__logger.debug("Loading cache for Kraken CSV pair converter.")
            self.__load_cache()

        # Attempt to load smallest duration
        if self.__cached_pairs.get(base_asset + quote_asset + "1"):
            self.__logger.debug("Retrieving cached bar for %s, %s at %s", base_asset, quote_asset, epoch_timestamp)
            return self._retrieve_cached_bar(base_asset, quote_asset, epoch_timestamp)

        base_file: str = f"{base_asset}_OHLCVT.zip"

        self.__logger.info("Attempting to load %s from Kraken Google Drive.", base_file)
        file_bytes = self._google_file_to_bytes(base_file)

        if file_bytes:
            with ZipFile(BytesIO(file_bytes)) as zipped_ohlcvt:
                self.__logger.debug("Files found in zipped file - %s", zipped_ohlcvt.namelist())
                all_timespans_for_pair: List[str] = [x for x in zipped_ohlcvt.namelist() if x.startswith(f"{base_asset}{quote_asset}_")]
                if len(all_timespans_for_pair) == 0:
                    self.__logger.debug("Market not found in Kraken files. Skipping file read.")
                    return None

                csv_files: Dict[str, str] = {}
                for file_name in all_timespans_for_pair:
                    self.__logger.debug("Reading in file %s for Kraken CSV pricing.", file_name)
                    csv_files[file_name] = zipped_ohlcvt.read(file_name).decode(encoding="utf-8")

                with ThreadPool(self.__THREAD_COUNT) as pool:
                    pool.starmap(self._split_chunks_size_n, zip(list(csv_files.keys()), list(csv_files.values())))

            save_to_cache(self.cache_key(), self.__cached_pairs)
            return self._retrieve_cached_bar(base_asset, quote_asset, epoch_timestamp)

        return None

    # isolated in order to be mocked
    def _google_file_to_bytes(self, file_name: str) -> Optional[bytes]:
        params: Dict[str, Any] = {
            _QUERY: f"'{_KRAKEN_FOLDER_ID}' in parents and name = '{file_name}'",
            _API_KEY: self.__google_api_key,
        }
        try:
            # Searching the Kraken folder for the specific file for the asset we are interested in
            # This query returns a JSON with the file ID we need to download the specific file we need.
            response: Response = self.__session.get(_GOOGLE_APIS_URL, params=params, timeout=self.__TIMEOUT)
            # {
            #  "kind": "drive#fileList",
            #  "incompleteSearch": false,
            #  "files": [
            #   {
            #    "kind": "drive#file",
            #    "id": "1AAWkwfxJjOvZQKv3c5XOH1ZjoIQMblQt",
            #    "name": "USDT_OHLCVT.zip",
            #    "mimeType": "application/zip"
            #   }
            #  ]
            # }

            ## Error Response - Key is invalid
            # {
            #  'error': {
            #    'code': 400,
            #    'message': 'API key not valid. Please pass a valid API key.',
            #    'errors': [
            #      {
            #        'message': 'API key not valid. Please pass a valid API key.',
            #        'domain': 'global',
            #        'reason': 'badRequest'
            #      }
            #    ],
            #    'status': 'INVALID_ARGUMENT',
            #    'details': [
            #      {
            #        '@type': 'type.googleapis.com/google.rpc.ErrorInfo',
            #        'reason': 'API_KEY_INVALID',
            #        'domain': 'googleapis.com',
            #        'metadata': {
            #          'service': 'drive.googleapis.com'
            #        }
            #      }
            #    ]
            #  }
            # } from https://www.googleapis.com/drive/v3/files?q=...

            ## Error Response - Key is valid but Google Drive API is specifically not configured
            # {
            #  'error': {
            #   'errors': [
            #    {
            #     'domain': 'usageLimits',
            #     'reason': 'accessNotConfigured',
            #     'message': 'Access Not Configured...',
            #     'extendedHelp': 'https://console.developers.google.com/apis/api/drive.googleapis.com/overview?project=728279122903'
            #    }
            #   ],
            #   'code': 403,
            #   'message': 'Access Not Configured...'
            #  }
            # }
            data: Any = response.json()

            error_json: Any = data.get(_ERROR)

            if error_json is not None:
                errors: Any = error_json[_ERRORS]
                for error in errors:
                    if error[_REASON] == _ACCESS_NOT_CONFIGURED:
                        self.__logger.error(
                            "Access not granted to Google Drive API. You must grant authorization to the Google"
                            "Drive API for your API key. Follow the link in the message for more details. Message:\n%s",
                            error[_MESSAGE],
                        )
                        raise RP2RuntimeError("Google Drive not authorized")
                    if error[_REASON] == _BAD_REQUEST:
                        self.__logger.error(
                            "Google API key is invalid. Please check that you have entered the key correctly and try again."
                            "If the problem persists, you can leave the field blank to use the REST API.\n%s",
                            error[_MESSAGE],
                        )
                        raise RP2RuntimeError("Google Drive key invalid")
            if not data.get(_FILES):
                self.__logger.debug("The file '%s' was not found on the Kraken Google Drive.")
                return None

            self.__logger.debug("Retrieved %s from %s", data, response.url)

            # Downloading the zipfile that contains the 6 files one for each of the standard durations of candles:
            # 1m, 5m, 15m, 1h, 12h, 24h.
            params = {_ALT: _MEDIA, _API_KEY: self.__google_api_key, _CONFIRM: 1}  # _CONFIRM: 1 bypasses large file warning
            file_response: Response = self.__session.get(f"{_GOOGLE_APIS_URL}/{data[_FILES][0][_ID]}", params=params, timeout=self.__TIMEOUT)

        except JSONDecodeError as exc:
            self.__logger.debug("Fetching of kraken csv files failed. Try again later.")
            raise RP2RuntimeError("JSON decode error") from exc

        return file_response.content



# Copyright 2022 eprbell
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from datetime import datetime, timedelta, timezone
from typing import Any

from rp2.rp2_decimal import RP2Decimal

from dali.abstract_pair_converter_plugin import AssetPairAndTimestamp
from dali.cache import CACHE_DIR, load_from_cache
from dali.configuration import Keyword
from dali.historical_bar import HistoricalBar
from dali.plugin.pair_converter.historic_crypto import PairConverterPlugin

BAR_DURATION: timedelta = timedelta(seconds=60)
BAR_TIMESTAMP: datetime = datetime(2020, 6, 1, 0, 0).replace(tzinfo=timezone.utc)
BAR_LOW: RP2Decimal = RP2Decimal("9430.01")
BAR_HIGH: RP2Decimal = RP2Decimal("9447.52")
BAR_OPEN: RP2Decimal = RP2Decimal("9445.83")
BAR_CLOSE: RP2Decimal = RP2Decimal("9435.80")
BAR_VOLUME: RP2Decimal = RP2Decimal("1")


class TestHistoricCryptoPlugin:
    def test_historical_prices(self, mocker: Any) -> None:
        plugin: PairConverterPlugin = PairConverterPlugin(Keyword.HISTORICAL_PRICE_HIGH.value)
        cache_path = os.path.join(CACHE_DIR, plugin.cache_key())
        if os.path.exists(cache_path):
            os.remove(cache_path)

        # Reinstantiate plugin now that cache is gone
        plugin = PairConverterPlugin(Keyword.HISTORICAL_PRICE_HIGH.value)
        mocker.patch.object(plugin, "get_historic_bar_from_native_source").return_value = HistoricalBar(
            duration=BAR_DURATION,
            timestamp=BAR_TIMESTAMP,
            low=BAR_LOW,
            high=BAR_HIGH,
            open=BAR_OPEN,
            close=BAR_CLOSE,
            volume=BAR_VOLUME,
        )

        # Read price without cache
        data = plugin.get_historic_bar_from_native_source(BAR_TIMESTAMP, "BTC", "USD", "Coinbase")

        assert data
        assert data.timestamp == BAR_TIMESTAMP
        assert data.timestamp == BAR_TIMESTAMP
        assert data.low == BAR_LOW
        assert data.high == BAR_HIGH
        assert data.open == BAR_OPEN
        assert data.close == BAR_CLOSE
        assert data.volume == BAR_VOLUME

        # Read price again, but populate plugin cache this time
        value = plugin.get_conversion_rate(BAR_TIMESTAMP, "BTC", "USD", "Coinbase")
        assert value
        assert value == BAR_HIGH

        # Save plugin cache
        plugin.save_historical_price_cache()

        # Load plugin cache and verify
        cache = load_from_cache(plugin.cache_key())
        key = AssetPairAndTimestamp(BAR_TIMESTAMP, "BTC", "USD", "Coinbase")
        assert len(cache) == 1, str(cache)
        assert key in cache
        data = cache[key]

        assert data
        assert data.timestamp == BAR_TIMESTAMP
        assert data.timestamp == BAR_TIMESTAMP
        assert data.low == BAR_LOW
        assert data.high == BAR_HIGH
        assert data.open == BAR_OPEN
        assert data.close == BAR_CLOSE
        assert data.volume == BAR_VOLUME

    def test_missing_historical_prices(self, mocker: Any) -> None:
        plugin = PairConverterPlugin(Keyword.HISTORICAL_PRICE_HIGH.value)
        timestamp = datetime(2020, 6, 1, 0, 0)

        mocker.patch.object(plugin, "get_historic_bar_from_native_source").return_value = None

        data = plugin.get_historic_bar_from_native_source(timestamp, "EUR", "JPY", "Coinbase")
        assert data is None



# Copyright 2022 macanudo527
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# pylint: disable=protected-access

import datetime
from itertools import chain, repeat
from typing import Any, Dict, List, Union

from ccxt import Exchange
from dateutil import parser

from rp2.plugin.country.us import US
from rp2.rp2_decimal import RP2Decimal

from dali.configuration import Keyword
from dali.in_transaction import InTransaction
from dali.intra_transaction import IntraTransaction
from dali.out_transaction import OutTransaction
from dali.plugin.input.rest.binance_com import InputPlugin

# Time Constants
_MS_IN_DAY: int = 86400000
_MS_IN_HOUR: int = 3600000

_NORMAL_SUBSCRIPTION_TIME: int = 1552089166000
_EARLY_REDEEM_SUBSCRIPTION_TIME: int = 1552089166000 + _MS_IN_DAY


class TestBinance:
    def test_deposits(self, mocker: Any) -> None:
        plugin = InputPlugin(
            account_holder="tester",
            api_key="a",
            api_secret="b",
            native_fiat="USD",
        )

        client: Exchange = plugin._client

        mocker.patch.object(client, "fetch_markets").return_value = [{"id": "ETHBTC"}]
        mocker.patch.object(plugin, "_AbstractCcxtInputPlugin__start_time_ms", int(datetime.datetime.now().timestamp()) * 1000 - 1)
        mocker.patch.object(client, "fetch_deposits").return_value = [
            {
                "info": {
                    "amount": "0.00999800",
                    "coin": "PAXG",
                    "network": "ETH",
                    "status": "1",
                    "address": "0x788cabe9236ce061e5a892e1a59395a81fc8d62c",
                    "addressTag": "",
                    "txId": "0xaad4654a3234aa6118af9b4b335f5ae81c360b2394721c019b5d1e75328b09f3",
                    "insertTime": "1599621997000",
                    "transferType": "0",
                    "confirmTimes": "12/12",
                    "unlockConfirm": "12/12",
                    "walletType": "0",
                },
                "id": None,
                "txid": "0xaad4654a3234aa6118af9b4b335f5ae81c360b2394721c019b5d1e75328b09f3",
                "timestamp": 1599621997000,
                "datetime": "2020-09-09T03:26:37.000Z",
                "network": "ETH",
                "address": "0x788cabe9236ce061e5a892e1a59395a81fc8d62c",
                "addressTo": "0x788cabe9236ce061e5a892e1a59395a81fc8d62c",
                "addressFrom": None,
                "tag": None,
                "tagTo": None,
                "tagFrom": None,
                "type": "deposit",
                "amount": 0.00999800,
                "currency": "PAXG",
                "status": "ok",
                "updated": None,
                "internal": False,
                "fee": None,
            }
        ]

        mocker.patch.object(plugin, "_process_trades").return_value = None
        mocker.patch.object(plugin, "_process_gains").return_value = None
        mocker.patch.object(plugin, "_process_withdrawals").return_value = None
        mocker.patch.object(plugin, "_process_implicit_api").return_value = None

        result = plugin.load(US())

        # 1 crypto Transfer
        assert len(result) == 1

        crypto_deposit_transaction: IntraTransaction = result[0]  # type: ignore

        assert crypto_deposit_transaction.asset == "PAXG"
        assert int(parser.parse(crypto_deposit_transaction.timestamp).timestamp()) * 1000 == 1599621997000
        assert crypto_deposit_transaction.from_exchange == Keyword.UNKNOWN.value
        assert crypto_deposit_transaction.to_exchange == "Binance.com"
        assert crypto_deposit_transaction.crypto_sent == Keyword.UNKNOWN.value
        assert RP2Decimal(crypto_deposit_transaction.crypto_received) == RP2Decimal("0.00999800")

    def test_trades(self, mocker: Any) -> None:
        plugin = InputPlugin(
            account_holder="tester",
            api_key="a",
            api_secret="b",
            native_fiat="GBP",
        )

        client: Exchange = plugin._client

        mocker.patch.object(client, "fetch_markets").return_value = [{"id": "ETHBTC", "type": "spot"}]
        # itertools.chain / repeat allows us to return one value once, and empty values after that
        mocker.patch.object(client, "fetch_my_trades").side_effect = chain(
            [
                [
                    # Trade using BNB for fee payment
                    {
                        "info": {"sample": "data"},  # the original decoded JSON as is
                        "id": "12345-67890:09876/54321",  # string trade id
                        "timestamp": 1502962946000,  # Unix timestamp in milliseconds
                        "datetime": "2017-08-17 12:42:48.000",  # ISO8601 datetime with milliseconds
                        "symbol": "ETH/BTC",  # symbol
                        "order": "12345-67890:09876/54321",  # string order id or undefined/None/null
                        "type": "limit",  # order type, 'market', 'limit' or undefined/None/null
                        "side": "buy",  # direction of the trade, 'buy' or 'sell'
                        "takerOrMaker": "taker",  # string, 'taker' or 'maker'
                        "price": 0.06917684,  # float price in quote currency
                        "amount": 1.5,  # amount of base currency
                        "cost": 0.10376526,  # total cost, `price * amount`,
                        "fee": {  # provided by exchange or calculated by ccxt
                            "cost": 0.0015,  # float
                            "currency": "BNB",  # usually base currency for buys, quote currency for sells
                            "rate": 0.002,  # the fee rate (if available)
                        },
                    },
                    # Trade using the quote currency for fee payment
                    {
                        "info": {"sample": "data"},  # the original decoded JSON as is
                        "id": "12345-67890:09876/54321",  # string trade id
                        "timestamp": 1502962947000,  # Unix timestamp in milliseconds
                        "datetime": "2017-08-17 12:42:48.000",  # ISO8601 datetime with milliseconds
                        "symbol": "ETH/BTC",  # symbol
                        "order": "12345-67890:09876/54321",  # string order id or undefined/None/null
                        "type": "limit",  # order type, 'market', 'limit' or undefined/None/null
                        "side": "buy",  # direction of the trade, 'buy' or 'sell'
                        "takerOrMaker": "taker",  # string, 'taker' or 'maker'
                        "price": 0.06917684,  # float price in quote currency
                        "amount": 3,  # amount of base currency
                        "cost": 0.20753052,  # total cost, `price * amount`,
                        "fee": {  # provided by exchange or calculated by ccxt
                            "cost": 0.0015,  # float
                            "currency": "ETH",  # usually base currency for buys, quote currency for sells
                            "rate": 0.002,  # the fee rate (if available)
                        },
                    },
                    # Sell trade using the quote currency for fee payment
                    {
                        "info": {"sample": "data"},  # the original decoded JSON as is
                        "id": "12345-67890:09876/54321",  # string trade id
                        "timestamp": 1502962948000,  # Unix timestamp in milliseconds
                        "datetime": "2017-08-17 12:42:48.000",  # ISO8601 datetime with milliseconds
                        "symbol": "ETH/BTC",  # symbol
                        "order": "12345-67890:09876/54321",  # string order id or undefined/None/null
                        "type": "limit",  # order type, 'market', 'limit' or undefined/None/null
                        "side": "sell",  # direction of the trade, 'buy' or 'sell'
                        "takerOrMaker": "taker",  # string, 'taker' or 'maker'
                        "price": 0.06917684,  # float price in quote currency
                        "amount": 6,  # amount of base currency
                        "cost": 0.41506104,  # total cost, `price * amount`,
                        "fee": {  # provided by exchange or calculated by ccxt
                            "cost": 0.0015,  # float
                            "currency": "BTC",  # usually base currency for buys, quote currency for sells
                            "rate": 0.002,  # the fee rate (if available)
                        },
                    },
                    # Fiat Buy
                    {
                        "info": {"sample": "data"},  # the original decoded JSON as is
                        "id": "12345-67890:09876/54321",  # string trade id
                        "timestamp": 1502962949000,  # Unix timestamp in milliseconds
                        "datetime": "2017-08-17 12:42:48.000",  # ISO8601 datetime with milliseconds
                        "symbol": "BTC/GBP",  # symbol
                        "order": "12345-67890:09876/54321",  # string order id or undefined/None/null
                        "type": "limit",  # order type, 'market', 'limit' or undefined/None/null
                        "side": "buy",  # direction of the trade, 'buy' or 'sell'
                        "takerOrMaker": "taker",  # string, 'taker' or 'maker'
                        "price": 23000.01,  # float price in quote currency
                        "amount": 1,  # amount of base currency
                        "cost": 23000.01,  # total cost, `price * amount`,
                        "fee": {  # provided by exchange or calculated by ccxt
                            "cost": 0.002,  # float
                            "currency": "BTC",  # usually base currency for buys, quote currency for sells
                            "rate": 0.002,  # the fee rate (if available)
                        },
                    },
                    # Fiat Sell
                    {
                        "info": {"sample": "data"},  # the original decoded JSON as is
                        "id": "12345-67890:09876/54321",  # string trade id
                        "timestamp": 1502962950000,  # Unix timestamp in milliseconds
                        "datetime": "2017-08-17 12:42:48.000",  # ISO8601 datetime with milliseconds
                        "symbol": "BTC/GBP",  # symbol
                        "order": "12345-67890:09876/54321",  # string order id or undefined/None/null
                        "type": "limit",  # order type, 'market', 'limit' or undefined/None/null
                        "side": "sell",  # direction of the trade, 'buy' or 'sell'
                        "takerOrMaker": "taker",  # string, 'taker' or 'maker'
                        "price": 23000.01,  # float price in quote currency
                        "amount": 1,  # amount of base currency
                        "cost": 23000.01,  # total cost, `price * amount`,
                        "fee": {  # provided by exchange or calculated by ccxt
                            "cost": 40,  # float
                            "currency": "GBP",  # usually base currency for buys, quote currency for sells
                            "rate": 0.002,  # the fee rate (if available)
                        },
                    },
                ]
            ],
            repeat([]),
        )

        # CCXT abstracts dust trades into regular trades, so no testing is necessary
        mocker.patch.object(client, "fetch_my_dust_trades").return_value = []
        mocker.patch.object(plugin, "_process_deposits").return_value = None
        mocker.patch.object(plugin, "_process_gains").return_value = None
        mocker.patch.object(plugin, "_process_withdrawals").return_value = None
        mocker.patch.object(plugin, "_process_implicit_api").return_value = None

        result = plugin.load(US())

        # One Sell of quote asset (using BNB) +
        # One Buy of base asset (using BNB) +
        # One payment for fees in BNB
        # One Sell of quote asset (for Buy order) +
        # One Buy of base asset (for Buy order) +
        # One Sell of quote asset (for Sell order) +
        # One Buy of base asset (for Sell order) +
        # One Buy of base asset (for fiat buy) +
        # One Sell of quote asset (for fiat buy) +
        # One Sell of quote asset (for fiat sell) +
        # One Buy of base asset (for fiat buy) = 11
        assert len(result) == 11

        bnb_sell_transaction: OutTransaction = result[6]  # type: ignore
        bnb_buy_transaction: InTransaction = result[0]  # type: ignore
        bnb_fee_transaction: OutTransaction = result[5]  # type: ignore
        buy_conversion_out: OutTransaction = result[7]  # type: ignore
        buy_conversion_in: InTransaction = result[1]  # type: ignore
        sell_conversion_order_out: OutTransaction = result[8]  # type: ignore
        sell_conversion_order_in: InTransaction = result[2]  # type: ignore
        buy_fiat_order_out: OutTransaction = result[9]  # type: ignore
        buy_fiat_order_in: InTransaction = result[3]  # type: ignore
        sell_fiat_order_out: OutTransaction = result[10]  # type: ignore
        sell_fiat_order_in: InTransaction = result[4]  # type: ignore

        # Buy with BNB as fee payment
        assert bnb_sell_transaction.asset == "BTC"
        assert int(parser.parse(bnb_sell_transaction.timestamp).timestamp()) * 1000 == 1502962946000
        assert bnb_sell_transaction.transaction_type == Keyword.SELL.value.capitalize()
        assert bnb_sell_transaction.spot_price == Keyword.UNKNOWN.value
        assert RP2Decimal(bnb_sell_transaction.crypto_out_no_fee) == RP2Decimal("0.10376526")
        assert RP2Decimal(bnb_sell_transaction.crypto_fee) == RP2Decimal("0")
        assert RP2Decimal(str(bnb_sell_transaction.crypto_out_with_fee)) == RP2Decimal("0.10376526")
        assert bnb_sell_transaction.fiat_out_no_fee is None
        assert bnb_sell_transaction.fiat_fee is None

        assert bnb_buy_transaction.asset == "ETH"
        assert int(parser.parse(bnb_buy_transaction.timestamp).timestamp()) * 1000 == 1502962946000
        assert bnb_buy_transaction.transaction_type == Keyword.BUY.value.capitalize()
        assert bnb_buy_transaction.spot_price == Keyword.UNKNOWN.value
        assert RP2Decimal(bnb_buy_transaction.crypto_in) == RP2Decimal("1.5")
        assert RP2Decimal(str(bnb_buy_transaction.crypto_fee)) == RP2Decimal("0")
        assert bnb_buy_transaction.fiat_in_no_fee is None
        assert bnb_buy_transaction.fiat_in_with_fee is None
        assert bnb_buy_transaction.fiat_fee is None

        assert bnb_fee_transaction.asset == "BNB"
        assert int(parser.parse(bnb_fee_transaction.timestamp).timestamp()) * 1000 == 1502962946000
        assert bnb_fee_transaction.transaction_type == Keyword.FEE.value.capitalize()
        assert bnb_fee_transaction.spot_price == Keyword.UNKNOWN.value
        assert RP2Decimal(bnb_fee_transaction.crypto_out_no_fee) == RP2Decimal("0")
        assert RP2Decimal(bnb_fee_transaction.crypto_fee) == RP2Decimal("0.0015")
        assert RP2Decimal(str(bnb_fee_transaction.crypto_out_with_fee)) == RP2Decimal("0.0015")
        assert bnb_fee_transaction.fiat_out_no_fee is None
        assert bnb_fee_transaction.fiat_fee is None

        # Buy with base asset as fee payment
        assert buy_conversion_out.asset == "BTC"
        assert int(parser.parse(buy_conversion_out.timestamp).timestamp()) * 1000 == 1502962947000
        assert buy_conversion_out.transaction_type == Keyword.SELL.value.capitalize()
        assert buy_conversion_out.spot_price == Keyword.UNKNOWN.value
        assert RP2Decimal(buy_conversion_out.crypto_out_no_fee) == RP2Decimal("0.20753052")
        assert RP2Decimal(buy_conversion_out.crypto_fee) == RP2Decimal("0")
        assert RP2Decimal(str(buy_conversion_out.crypto_out_with_fee)) == RP2Decimal("0.20753052")
        assert buy_conversion_out.fiat_out_no_fee is None
        assert buy_conversion_out.fiat_fee is None

        assert buy_conversion_in.asset == "ETH"
        assert int(parser.parse(buy_conversion_in.timestamp).timestamp()) * 1000 == 1502962947000
        assert buy_conversion_in.transaction_type == Keyword.BUY.value.capitalize()
        assert buy_conversion_in.spot_price == Keyword.UNKNOWN.value
        assert RP2Decimal(buy_conversion_in.crypto_in) == RP2Decimal("3")
        assert RP2Decimal(str(buy_conversion_in.crypto_fee)) == RP2Decimal("0.0015")
        assert buy_conversion_in.fiat_in_no_fee is None
        assert buy_conversion_in.fiat_in_with_fee is None
        assert buy_conversion_in.fiat_fee is None

        # Sell with quote asset as fee payment
        assert sell_conversion_order_out.asset == "ETH"
        assert int(parser.parse(sell_conversion_order_out.timestamp).timestamp()) * 1000 == 1502962948000
        assert sell_conversion_order_out.transaction_type == Keyword.SELL.value.capitalize()
        assert sell_conversion_order_out.spot_price == Keyword.UNKNOWN.value
        assert RP2Decimal(sell_conversion_order_out.crypto_out_no_fee) == RP2Decimal("6")
        assert RP2Decimal(sell_conversion_order_out.crypto_fee) == RP2Decimal("0")
        assert RP2Decimal(str(sell_conversion_order_out.crypto_out_with_fee)) == RP2Decimal("6")
        assert sell_conversion_order_out.fiat_out_no_fee is None
        assert sell_conversion_order_out.fiat_fee is None

        assert sell_conversion_order_in.asset == "BTC"
        assert int(parser.parse(sell_conversion_order_in.timestamp).timestamp()) * 1000 == 1502962948000
        assert sell_conversion_order_in.transaction_type == Keyword.BUY.value.capitalize()
        assert sell_conversion_order_in.spot_price == Keyword.UNKNOWN.value
        assert RP2Decimal(sell_conversion_order_in.crypto_in) == RP2Decimal("0.41506104")
        assert RP2Decimal(str(sell_conversion_order_in.crypto_fee)) == RP2Decimal("0.0015")
        assert sell_conversion_order_in.fiat_in_no_fee is None
        assert sell_conversion_order_in.fiat_in_with_fee is None
        assert sell_conversion_order_in.fiat_fee is None

        # Fiat buy with base asset as a fee
        assert buy_fiat_order_out.asset == "GBP"
        assert int(parser.parse(buy_fiat_order_out.timestamp).timestamp()) * 1000 == 1502962949000
        assert buy_fiat_order_out.transaction_type == Keyword.SELL.value.capitalize()
        assert RP2Decimal(buy_fiat_order_out.spot_price) == RP2Decimal("23000.01")
        assert RP2Decimal(buy_fiat_order_out.crypto_out_no_fee) == RP2Decimal("23000.01")
        assert RP2Decimal(buy_fiat_order_out.crypto_fee) == RP2Decimal("0")
        assert RP2Decimal(str(buy_fiat_order_out.crypto_out_with_fee)) == RP2Decimal("23000.01")
        assert RP2Decimal(str(buy_fiat_order_out.fiat_out_no_fee)) == RP2Decimal("23000.01")
        assert RP2Decimal(str(buy_fiat_order_out.fiat_fee)) == RP2Decimal("0")
        assert buy_fiat_order_out.fiat_ticker == "GBP"

        assert buy_fiat_order_in.asset == "BTC"
        assert int(parser.parse(buy_fiat_order_in.timestamp).timestamp()) * 1000 == 1502962949000
        assert buy_fiat_order_in.transaction_type == Keyword.BUY.value.capitalize()
        assert RP2Decimal(buy_fiat_order_in.spot_price) == RP2Decimal("23000.01")
        assert RP2Decimal(buy_fiat_order_in.crypto_in) == RP2Decimal("1")
        assert RP2Decimal(str(buy_fiat_order_in.crypto_fee)) == RP2Decimal("0.002")
        assert RP2Decimal(str(buy_fiat_order_in.fiat_in_no_fee)) == RP2Decimal("22954.00998")
        assert RP2Decimal(str(buy_fiat_order_in.fiat_in_with_fee)) == RP2Decimal("23000.01")
        assert buy_fiat_order_in.fiat_fee is None

        # Fiat sell with quote asset as a fee
        assert sell_fiat_order_out.asset == "BTC"
        assert int(parser.parse(sell_fiat_order_out.timestamp).timestamp()) * 1000 == 1502962950000
        assert sell_fiat_order_out.transaction_type == Keyword.SELL.value.capitalize()
        assert RP2Decimal(sell_fiat_order_out.spot_price) == RP2Decimal("23000.01")
        assert RP2Decimal(sell_fiat_order_out.crypto_out_no_fee) == RP2Decimal("1")
        assert RP2Decimal(sell_fiat_order_out.crypto_fee) == RP2Decimal("0")
        assert RP2Decimal(str(sell_fiat_order_out.crypto_out_with_fee)) == RP2Decimal("1")
        assert RP2Decimal(str(sell_fiat_order_out.fiat_out_no_fee)) == RP2Decimal("23000.01")
        assert RP2Decimal(str(sell_fiat_order_out.fiat_fee)) == RP2Decimal("0")

        assert sell_fiat_order_in.asset == "GBP"
        assert int(parser.parse(sell_fiat_order_in.timestamp).timestamp()) * 1000 == 1502962950000
        assert sell_fiat_order_in.transaction_type == Keyword.BUY.value.capitalize()
        assert RP2Decimal(sell_fiat_order_in.spot_price) == RP2Decimal("23000.01")
        assert RP2Decimal(sell_fiat_order_in.crypto_in) == RP2Decimal("23000.01")
        assert sell_fiat_order_in.crypto_fee is None
        assert RP2Decimal(str(sell_fiat_order_in.fiat_in_no_fee)) == RP2Decimal("22960.01")
        assert RP2Decimal(str(sell_fiat_order_in.fiat_in_with_fee)) == RP2Decimal("23000.01")
        assert RP2Decimal(str(sell_fiat_order_in.fiat_fee)) == RP2Decimal("40")
        assert sell_fiat_order_in.fiat_ticker == "GBP"

    def test_gains(self, mocker: Any) -> None:
        plugin = InputPlugin(
            account_holder="tester",
            api_key="a",
            api_secret="b",
            username="user",
            native_fiat="USD",
        )

        client: Exchange = plugin._client

        mocker.patch.object(client, "fetch_markets").return_value = [{"id": "ETHBTC"}]
        mocker.patch.object(client, "sapiGetMiningPubAlgoList").return_value = {"data": [{"algoName": "sha256"}]}
        mocker.patch.object(plugin, "_AbstractCcxtInputPlugin__start_time_ms", int(datetime.datetime.now().timestamp()) * 1000 - 1)
        mocker.patch.object(client, "sapiGetAssetAssetDividend").return_value = {
            "rows": [
                {"id": 1637366104, "amount": "0.00001600", "asset": "BETH", "divTime": 1563189166000, "enInfo": "ETH 2.0 Staking", "tranId": 2968885920},
                {"id": 1631750237, "amount": "0.51206985", "asset": "BUSD", "divTime": 1563189165000, "enInfo": "Flexible Savings", "tranId": 2968885920},
            ],
            "total": 2,
        }

        # Only mining type 0 transactions are supported
        mocker.patch.object(client, "sapiGetMiningPaymentList").return_value = {
            "code": 0,
            "msg": "",
            "data": {
                "accountProfits": [
                    {
                        "time": 1586188800000,  # Mining date
                        "type": "31",  # 0:Mining Wallet,5:Mining Address,7:Pool Savings,8:Transferred,31:Income Transfer ,
                        # 32:Hashrate Resale-Mining Wallet 33:Hashrate Resale-Pool Savings
                        "hashTransfer": None,  # Transferred Hashrate
                        "transferAmount": None,  # Transferred Income
                        "dayHashRate": 129129903378244,  # Daily Hashrate
                        "profitAmount": 8.6083060304,  # Earnings Amount
                        "coinName": "BTC",  # Coin Type
                        "status": "2",  # Status：0:Unpaid， 1:Paying  2：Paid
                    },
                    {
                        "time": 1607529600000,
                        "coinName": "BTC",
                        "type": "0",  # String
                        "dayHashRate": 9942053925926,
                        "profitAmount": 0.85426469,
                        "hashTransfer": 200000000000,
                        "transferAmount": 0.02180958,
                        "status": "2",
                    },
                ],
                "totalNum": 2,  # Total Rows
                "pageSize": 20,  # Rows per page
            },
        }

        def penalized_locked_staking(params: Dict[str, str]) -> List[Dict[str, str]]:
            # pylint: disable=unused-argument
            result: List[Dict[str, str]] = []

            if params["txnType"] == "SUBSCRIPTION":
                result = [
                    {
                        "time": str(_NORMAL_SUBSCRIPTION_TIME),
                        "asset": "BTC",
                        "amount": "1",
                        "lockPeriod": "10",
                        "type": "NORMAL",
                        "status": "SUCCESS",
                    },
                    {
                        "time": str(_EARLY_REDEEM_SUBSCRIPTION_TIME),
                        "asset": "BTC",
                        "amount": "1",
                        "lockPeriod": "10",
                        "type": "NORMAL",
                        "status": "SUCCESS",
                    },
                ]

            elif params["txnType"] == "REDEMPTION":
                result = [
                    {
                        "positionId": "12345",
                        "time": str(_NORMAL_SUBSCRIPTION_TIME + _MS_IN_DAY * 10 + _MS_IN_HOUR * 12),  # valid redemption time
                        "asset": "BTC",
                        "amount": "1",
                        "deliverDate": str(_NORMAL_SUBSCRIPTION_TIME + _MS_IN_DAY * 10 + _MS_IN_HOUR * 13),  # Just needs to be later
                        "status": "PAID",
                    },
                    {
                        "positionId": "12346",
                        "time": str(_EARLY_REDEEM_SUBSCRIPTION_TIME + _MS_IN_DAY * 5 + _MS_IN_HOUR * 12),  # invalid redemption time (6 days short)
                        "asset": "BTC",
                        "amount": "0.95",
                        "deliverDate": str(_EARLY_REDEEM_SUBSCRIPTION_TIME + _MS_IN_DAY * 5 + _MS_IN_HOUR * 13),  # Just needs to be later
                        "status": "PAID",
                    },
                ]
            elif params["txnType"] == "INTEREST":
                normal_interest_start: int = _NORMAL_SUBSCRIPTION_TIME + _MS_IN_HOUR * 6
                early_redem_interest_start: int = _EARLY_REDEEM_SUBSCRIPTION_TIME + _MS_IN_HOUR * 6
                for i in range(1, 11):
                    result.append(
                        {
                            "positionId": "12345",
                            "time": str(normal_interest_start + _MS_IN_DAY * i),
                            "asset": "BTC",
                            "amount": "0.01",
                            "status": "SUCCESS",
                        }
                    )

                    # Early redemption doesn't get all the interest payments
                    if i <= 5:
                        result.append(
                            {
                                "positionId": "12346",
                                "time": str(early_redem_interest_start + _MS_IN_DAY * i),
                                "asset": "BTC",
                                "amount": "0.01",
                                "status": "SUCCESS",
                            }
                        )

            return result

        mocker.patch.object(client, "sapi_get_staking_stakingrecord").side_effect = penalized_locked_staking

        mocker.patch.object(plugin, "_process_deposits").return_value = None
        mocker.patch.object(plugin, "_process_trades").return_value = None
        mocker.patch.object(plugin, "_process_withdrawals").return_value = None
        mocker.patch.object(plugin, "_process_implicit_api").return_value = None

        result = plugin.load(US())

        # One Eth staking transaction +
        # One BUSD savings transaction +
        # 15 Interest transactions +
        # 1 Early Redemption Penalty +
        # One Mining transaction = 19
        assert len(result) == 19

        eth_staking: InTransaction = result[0]  # type: ignore
        busd_savings: InTransaction = result[1]  # type: ignore
        early_redeem_fee: OutTransaction = result[18]  # type: ignore
        mining_deposit: InTransaction = result[17]  # type: ignore

        # Make sure it identifies this as staking income
        assert eth_staking.asset == "BETH"
        assert int(parser.parse(eth_staking.timestamp).timestamp()) * 1000 == 1563189166000
        assert eth_staking.transaction_type == Keyword.STAKING.value.capitalize()
        assert eth_staking.spot_price == Keyword.UNKNOWN.value
        assert RP2Decimal(eth_staking.crypto_in) == RP2Decimal("0.00001600")
        assert eth_staking.crypto_fee is None
        assert eth_staking.fiat_in_no_fee is None
        assert eth_staking.fiat_in_with_fee is None
        assert eth_staking.fiat_fee is None

        assert busd_savings.asset == "BUSD"
        assert int(parser.parse(busd_savings.timestamp).timestamp()) * 1000 == 1563189165000
        assert busd_savings.transaction_type == Keyword.INTEREST.value.capitalize()
        assert busd_savings.spot_price == Keyword.UNKNOWN.value
        assert RP2Decimal(busd_savings.crypto_in) == RP2Decimal("0.51206985")
        assert busd_savings.crypto_fee is None
        assert busd_savings.fiat_in_no_fee is None
        assert busd_savings.fiat_in_with_fee is None
        assert busd_savings.fiat_fee is None

        assert early_redeem_fee.asset == "BTC"
        assert int(parser.parse(early_redeem_fee.timestamp).timestamp()) * 1000 == _EARLY_REDEEM_SUBSCRIPTION_TIME + _MS_IN_DAY * 5 + _MS_IN_HOUR * 13
        assert early_redeem_fee.transaction_type == Keyword.FEE.value.capitalize()
        assert early_redeem_fee.spot_price == Keyword.UNKNOWN.value
        assert RP2Decimal(early_redeem_fee.crypto_out_no_fee) == RP2Decimal("0")
        assert RP2Decimal(str(early_redeem_fee.crypto_out_with_fee)) == RP2Decimal("0.05")
        assert RP2Decimal(early_redeem_fee.crypto_fee) == RP2Decimal("0.05")
        assert early_redeem_fee.fiat_out_no_fee is None
        assert early_redeem_fee.fiat_fee is None

        assert mining_deposit.asset == "BTC"
        assert int(parser.parse(mining_deposit.timestamp).timestamp()) * 1000 == 1607529600000
        assert mining_deposit.transaction_type == Keyword.MINING.value.capitalize()
        assert mining_deposit.spot_price == Keyword.UNKNOWN.value
        assert RP2Decimal(mining_deposit.crypto_in) == RP2Decimal("0.85426469")
        assert mining_deposit.crypto_fee is None
        assert mining_deposit.fiat_in_no_fee is None
        assert mining_deposit.fiat_in_with_fee is None
        assert mining_deposit.fiat_fee is None

    def test_implicit_api(self, mocker: Any) -> None:
        plugin = InputPlugin(
            account_holder="tester",
            api_key="a",
            api_secret="b",
            native_fiat="GBP",
        )

        client = plugin._client

        mocker.patch.object(client, "fetch_markets").return_value = [{"id": "ETHBTC", "type": "spot"}]
        mocker.patch.object(plugin, "_AbstractCcxtInputPlugin__start_time_ms", int(datetime.datetime.now().timestamp()) * 1000 - 1)
        mocker.patch.object(client, "sapiGetFiatPayments").return_value = {
            "code": "000000",
            "message": "success",
            "data": [
                {
                    "orderNo": "353fca443f06466db0c4dc89f94f027a",
                    "sourceAmount": "20.0",  # Fiat trade amount
                    "fiatCurrency": "EUR",  # Fiat token
                    "obtainAmount": "4.462",  # Crypto trade amount
                    "cryptoCurrency": "LUNA",  # Crypto token
                    "totalFee": "0.2",  # Trade fee
                    "price": "4.437472",
                    "status": "Completed",  # Processing, Completed, Failed, Refunded
                    "createTime": 1624529919000,
                    "updateTime": 1624529919000,
                },
                {
                    "orderNo": "353fca443f06466db0c4dc89f94f027c",
                    "sourceAmount": "15.0",  # Fiat trade amount
                    "fiatCurrency": "GBP",  # Fiat token
                    "obtainAmount": "4.362",  # Crypto trade amount
                    "cryptoCurrency": "LUNA",  # Crypto token
                    "totalFee": "0.15",  # Trade fee
                    "price": "3.4044016506",
                    "status": "Completed",  # Processing, Completed, Failed, Refunded
                    "createTime": 1624529915000,
                    "updateTime": 1624529915000,
                },
                {
                    "orderNo": "353fca443f06466db0c4dc89f94f027b",
                    "sourceAmount": "40.0",  # Fiat trade amount
                    "fiatCurrency": "EUR",  # Fiat token
                    "obtainAmount": "8.924",  # Crypto trade amount
                    "cryptoCurrency": "LUNA",  # Crypto token
                    "totalFee": "0.4",  # Trade fee
                    "price": "4.437472",
                    "status": "Failed",  # Processing, Completed, Failed, Refunded
                    "createTime": 1624529920000,
                    "updateTime": 1624529920000,
                },
            ],
            "total": 2,
            "success": True,
        }

        def fiat_deposit_withdrawal(params: Dict[str, Union[int, str]]) -> Dict[str, Union[bool, int, List[Dict[str, Union[int, str]]], str]]:
            # pylint: disable=unused-argument
            result: Dict[str, Union[bool, int, List[Dict[str, Union[int, str]]], str]] = {}

            # 0 - deposit
            if params["transactionType"] == 0:
                result = {
                    "code": "000000",
                    "message": "success",
                    "data": [
                        {
                            "orderNo": "25ced37075c1470ba8939d0df2316e23",
                            "fiatCurrency": "EUR",
                            "indicatedAmount": "15.00",
                            "amount": "14.80",
                            "totalFee": "0.20",
                            "method": "card",
                            "status": "Completed",
                            "createTime": 1627501026000,
                            "updateTime": 1627501027000,
                        },
                        {
                            "orderNo": "25ced37075c1470ba8939d0df2316e24",
                            "fiatCurrency": "EUR",
                            "indicatedAmount": "30.00",
                            "amount": "29.60",
                            "totalFee": "0.40",
                            "method": "card",
                            "status": "Failed",
                            "createTime": 1627501028000,
                            "updateTime": 1627501029000,
                        },
                    ],
                    "total": 2,
                    "success": True,
                }
            # 1 - withdrawal
            elif params["transactionType"] == 1:
                result = {
                    "code": "000000",
                    "message": "success",
                    "data": [
                        {
                            "orderNo": "25ced37075c1470ba8939d0df2316e46",
                            "fiatCurrency": "GBP",
                            "indicatedAmount": "15.00",
                            "amount": "14.80",
                            "totalFee": "0.20",
                            "method": "card",
                            "status": "Completed",
                            "createTime": 1627501036000,
                            "updateTime": 1627501037000,
                        },
                        {
                            "orderNo": "25ced37075c1470ba8939d0df2316e24",
                            "fiatCurrency": "GBP",
                            "indicatedAmount": "30.00",
                            "amount": "29.60",
                            "totalFee": "0.40",
                            "method": "card",
                            "status": "Failed",
                            "createTime": 1627501038000,
                            "updateTime": 1627501039000,
                        },
                    ],
                    "total": 2,
                    "success": True,
                }

            return result

        mocker.patch.object(client, "sapiGetFiatOrders").side_effect = fiat_deposit_withdrawal
        mocker.patch.object(client, "fetch_my_dust_trades").return_value = []
        mocker.patch.object(plugin, "_process_trades").return_value = None
        mocker.patch.object(plugin, "_process_gains").return_value = None
        mocker.patch.object(plugin, "_process_withdrawals").return_value = None
        mocker.patch.object(plugin, "_process_deposits").return_value = None

        result = plugin.load(US())

        # 1 EUR deposit +
        # 1 EUR sell (priced in native fiat) +
        # 1 EUR -> LUNA conversion +
        # 1 GBP deposit +
        # 1 GBP sell +
        # 1 GBP -> LUNA buy +
        # 1 fiat deposit +
        # 1 fiat withdrawal = 8
        assert len(result) == 8

        eur_deposit_transaction: InTransaction = next(
            item for item in result if item.unique_id == "353fca443f06466db0c4dc89f94f027a/fiat_buy" and item.asset == "EUR" and isinstance(item, InTransaction)
        )
        eur_sell_transaction: OutTransaction = next(
            item for item in result if item.unique_id == "353fca443f06466db0c4dc89f94f027a" and isinstance(item, OutTransaction)
        )
        eur_luna_conversion_transaction: InTransaction = next(
            item for item in result if item.unique_id == "353fca443f06466db0c4dc89f94f027a" and item.asset == "LUNA" and isinstance(item, InTransaction)
        )
        gbp_deposit_transaction: InTransaction = next(
            item for item in result if item.unique_id == "353fca443f06466db0c4dc89f94f027c/fiat_buy" and item.asset == "GBP" and isinstance(item, InTransaction)
        )
        gbp_sell_transaction: OutTransaction = next(
            item for item in result if item.unique_id == "353fca443f06466db0c4dc89f94f027c" and isinstance(item, OutTransaction)
        )
        gbp_luna_buy_transaction: InTransaction = next(
            item for item in result if item.unique_id == "353fca443f06466db0c4dc89f94f027c" and item.asset == "LUNA" and isinstance(item, InTransaction)
        )
        fiat_deposit: InTransaction = next(item for item in result if item.unique_id == "25ced37075c1470ba8939d0df2316e23" and isinstance(item, InTransaction))
        fiat_withdrawal: OutTransaction = next(
            item for item in result if item.unique_id == "25ced37075c1470ba8939d0df2316e46" and isinstance(item, OutTransaction)
        )

        # fiat deposit must still be recorded
        assert eur_deposit_transaction.asset == "EUR"
        assert int(parser.parse(eur_deposit_transaction.timestamp).timestamp()) * 1000 == 1624529919000
        assert eur_deposit_transaction.transaction_type == Keyword.BUY.value.capitalize()
        assert eur_deposit_transaction.spot_price == "1"
        assert RP2Decimal(eur_deposit_transaction.crypto_in) == RP2Decimal("20.0")
        assert eur_deposit_transaction.crypto_fee is None
        assert RP2Decimal(str(eur_deposit_transaction.fiat_in_no_fee)) == RP2Decimal("20.0")
        assert RP2Decimal(str(eur_deposit_transaction.fiat_in_with_fee)) == RP2Decimal("20.0")
        assert eur_deposit_transaction.fiat_fee is None
        assert eur_deposit_transaction.fiat_ticker == "EUR"

        # Purchase made in fiat other than native_fiat is a regular conversion
        assert eur_sell_transaction.asset == "EUR"
        assert int(parser.parse(eur_sell_transaction.timestamp).timestamp()) * 1000 == 1624529919000
        assert eur_sell_transaction.transaction_type == Keyword.SELL.value.capitalize()
        assert eur_sell_transaction.spot_price == Keyword.UNKNOWN.value
        assert RP2Decimal(eur_sell_transaction.crypto_out_no_fee) == RP2Decimal("19.8")
        assert RP2Decimal(eur_sell_transaction.crypto_fee) == RP2Decimal("0.2")
        assert eur_sell_transaction.fiat_out_no_fee is None
        assert eur_sell_transaction.fiat_fee is None

        assert eur_luna_conversion_transaction.asset == "LUNA"
        assert int(parser.parse(eur_luna_conversion_transaction.timestamp).timestamp()) * 1000 == 1624529919000
        assert eur_luna_conversion_transaction.transaction_type == Keyword.BUY.value.capitalize()
        assert eur_luna_conversion_transaction.spot_price == Keyword.UNKNOWN.value
        assert RP2Decimal(eur_luna_conversion_transaction.crypto_in) == RP2Decimal("4.462")
        assert eur_luna_conversion_transaction.crypto_fee is None
        assert eur_luna_conversion_transaction.fiat_in_no_fee is None
        assert eur_luna_conversion_transaction.fiat_in_with_fee is None
        assert eur_luna_conversion_transaction.fiat_fee is None

        assert gbp_deposit_transaction.asset == "GBP"
        assert int(parser.parse(gbp_deposit_transaction.timestamp).timestamp()) * 1000 == 1624529915000
        assert gbp_deposit_transaction.transaction_type == Keyword.BUY.value.capitalize()
        assert gbp_deposit_transaction.spot_price == "1"
        assert RP2Decimal(gbp_deposit_transaction.crypto_in) == RP2Decimal("15.0")
        assert gbp_deposit_transaction.crypto_fee is None
        assert RP2Decimal(str(gbp_deposit_transaction.fiat_in_no_fee)) == RP2Decimal("15.0")
        assert RP2Decimal(str(gbp_deposit_transaction.fiat_in_with_fee)) == RP2Decimal("15.0")
        assert gbp_deposit_transaction.fiat_fee is None
        assert gbp_deposit_transaction.fiat_ticker == "GBP"

        assert gbp_sell_transaction.asset == "GBP"
        assert int(parser.parse(gbp_sell_transaction.timestamp).timestamp()) * 1000 == 1624529915000
        assert gbp_sell_transaction.transaction_type == Keyword.SELL.value.capitalize()
        assert gbp_sell_transaction.spot_price == "1"
        assert RP2Decimal(gbp_sell_transaction.crypto_out_no_fee) == RP2Decimal("14.85")
        assert RP2Decimal(gbp_sell_transaction.crypto_fee) == RP2Decimal("0.15")
        assert RP2Decimal(str(gbp_sell_transaction.fiat_out_no_fee)) == RP2Decimal("14.85")
        assert gbp_sell_transaction.fiat_fee is None

        assert gbp_luna_buy_transaction.asset == "LUNA"
        assert int(parser.parse(gbp_luna_buy_transaction.timestamp).timestamp()) * 1000 == 1624529915000
        assert gbp_luna_buy_transaction.transaction_type == Keyword.BUY.value.capitalize()
        assert RP2Decimal(gbp_luna_buy_transaction.spot_price) == RP2Decimal("3.4044016506")
        assert RP2Decimal(gbp_luna_buy_transaction.crypto_in) == RP2Decimal("4.362")
        assert gbp_luna_buy_transaction.crypto_fee is None
        assert RP2Decimal(str(gbp_luna_buy_transaction.fiat_in_no_fee)) == RP2Decimal("14.85")
        assert RP2Decimal(str(gbp_luna_buy_transaction.fiat_in_with_fee)) == RP2Decimal("15.0")
        assert RP2Decimal(str(gbp_luna_buy_transaction.fiat_fee)) == RP2Decimal("0.15")
        assert gbp_luna_buy_transaction.fiat_ticker == "GBP"

        assert fiat_deposit.asset == "EUR"
        assert int(parser.parse(fiat_deposit.timestamp).timestamp()) * 1000 == 1627501026000
        assert fiat_deposit.transaction_type == Keyword.BUY.value.capitalize()
        assert RP2Decimal(fiat_deposit.spot_price) == RP2Decimal("1")
        assert RP2Decimal(fiat_deposit.crypto_in) == RP2Decimal("15.00")
        assert RP2Decimal(str(fiat_deposit.crypto_fee)) == RP2Decimal("0.20")
        assert fiat_deposit.fiat_in_no_fee is None
        assert fiat_deposit.fiat_in_with_fee is None
        assert fiat_deposit.fiat_fee is None
        assert fiat_deposit.fiat_ticker == "EUR"

        assert fiat_withdrawal.asset == "GBP"
        assert int(parser.parse(fiat_withdrawal.timestamp).timestamp()) * 1000 == 1627501036000
        assert fiat_withdrawal.transaction_type == Keyword.SELL.value.capitalize()
        assert RP2Decimal(fiat_withdrawal.spot_price) == RP2Decimal("1")
        assert RP2Decimal(fiat_withdrawal.crypto_out_no_fee) == RP2Decimal("15.00")
        assert RP2Decimal(str(fiat_withdrawal.crypto_fee)) == RP2Decimal("0.20")
        assert fiat_withdrawal.fiat_out_no_fee is None
        assert fiat_withdrawal.fiat_fee is None
        assert fiat_withdrawal.fiat_ticker == "GBP"

    def test_withdrawals(self, mocker: Any) -> None:
        plugin = InputPlugin(
            account_holder="tester",
            api_key="a",
            api_secret="b",
            native_fiat="USD",
        )

        client = plugin._client

        mocker.patch.object(client, "fetch_markets").return_value = [{"id": "ETHBTC"}]
        mocker.patch.object(plugin, "_AbstractCcxtInputPlugin__start_time_ms", int(datetime.datetime.now().timestamp()) * 1000 - 1)
        mocker.patch.object(client, "fetch_withdrawals").return_value = [
            {
                "info": {
                    "amount": "0.00999800",
                    "coin": "PAXG",
                    "network": "ETH",
                    "status": "1",
                    "address": "0x788cabe9236ce061e5a892e1a59395a81fc8d62c",
                    "addressTag": "",
                    "txId": "0xaad4654a3234aa6118af9b4b335f5ae81c360b2394721c019b5d1e75328b09f3",
                    "insertTime": "1599621997000",
                    "transferType": "0",
                    "confirmTimes": "12/12",
                    "unlockConfirm": "12/12",
                    "walletType": "0",
                },
                "id": None,
                "txid": "0xaad4654a3234aa6118af9b4b335f5ae81c360b2394721c019b5d1e75328b09f3",
                "timestamp": 1599621997000,
                "datetime": "2020-09-09T03:26:37.000Z",
                "network": "ETH",
                "address": "0x788cabe9236ce061e5a892e1a59395a81fc8d62c",
                "addressTo": "0x788cabe9236ce061e5a892e1a59395a81fc8d62c",
                "addressFrom": None,
                "tag": None,
                "tagTo": None,
                "tagFrom": None,
                "type": "withdrawal",
                "amount": 0.00999800,
                "currency": "PAXG",
                "status": "ok",
                "updated": None,
                "internal": False,
                "fee": None,
            }
        ]

        mocker.patch.object(plugin, "_process_trades").return_value = None
        mocker.patch.object(plugin, "_process_gains").return_value = None
        mocker.patch.object(plugin, "_process_deposits").return_value = None
        mocker.patch.object(plugin, "_process_implicit_api").return_value = None

        result = plugin.load(US())

        # 1 crypto Transfer
        assert len(result) == 1

        crypto_withdrawal_transaction: IntraTransaction = result[0]  # type: ignore

        assert crypto_withdrawal_transaction.asset == "PAXG"
        assert int(parser.parse(crypto_withdrawal_transaction.timestamp).timestamp()) * 1000 == 1599621997000
        assert crypto_withdrawal_transaction.to_exchange == Keyword.UNKNOWN.value
        assert crypto_withdrawal_transaction.from_exchange == "Binance.com"
        assert crypto_withdrawal_transaction.crypto_received == Keyword.UNKNOWN.value
        assert RP2Decimal(crypto_withdrawal_transaction.crypto_sent) == RP2Decimal("0.00999800")



# Copyright 2022 macanudo527
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Buys CSV Format: Internal ID, amount (amount received), (Total) Price, Trading Currency (Crypto Asset),
#    Original Currency (fiat used), Progress, UTC Timestamp

import logging
from csv import reader
from typing import List, Optional

from rp2.abstract_country import AbstractCountry
from rp2.logger import create_logger
from rp2.rp2_decimal import RP2Decimal

from dali.abstract_input_plugin import AbstractInputPlugin
from dali.abstract_transaction import AbstractTransaction
from dali.configuration import Keyword
from dali.in_transaction import InTransaction


class InputPlugin(AbstractInputPlugin):

    __COINCHECK: str = "Coincheck"
    __COINCHECK_SUPPLEMENTAL_PLUGIN: str = "Coincheck_Supplemental_CSV"

    __ID: int = 0
    __AMOUNT_PURCHASED: int = 1
    __TOTAL_PRICE: int = 2
    __CRYPTO_ASSET: int = 3
    __FIAT_USED: int = 4
    __PROGRESS: int = 5
    __TIMESTAMP_INDEX: int = 6

    __DELIMITER: str = ","

    def __init__(
        self,
        account_holder: str,
        buys_csv_file: str,
        native_fiat: Optional[str] = None,
    ) -> None:

        super().__init__(account_holder=account_holder, native_fiat=native_fiat)
        self.__buys_csv_file: str = buys_csv_file
        self.__logger: logging.Logger = create_logger(f"{self.__COINCHECK_SUPPLEMENTAL_PLUGIN}/{self.account_holder}")

    def load(self, country: AbstractCountry) -> List[AbstractTransaction]:
        return self.parse_buys_file(self.__buys_csv_file)

    def parse_buys_file(self, file_path: str) -> List[AbstractTransaction]:
        result: List[AbstractTransaction] = []

        with open(file_path, encoding="utf-8") as csv_file:
            lines = reader(csv_file)

            header = next(lines)
            self.__logger.debug("Header: %s", header)
            for line in lines:
                raw_data: str = self.__DELIMITER.join(line)
                self.__logger.debug("Transaction: %s", raw_data)

                unformatted_timestamp: str = line[self.__TIMESTAMP_INDEX]
                timestamp: str = unformatted_timestamp[: -len(" UTC")]

                result.append(
                    InTransaction(
                        plugin=self.__COINCHECK_SUPPLEMENTAL_PLUGIN,
                        unique_id=line[self.__ID],
                        raw_data=raw_data,
                        timestamp=f"{timestamp} -00:00",
                        asset=line[self.__CRYPTO_ASSET],
                        exchange=self.__COINCHECK,
                        holder=self.account_holder,
                        transaction_type=Keyword.BUY.value,
                        spot_price=str(RP2Decimal(line[self.__TOTAL_PRICE]) / RP2Decimal(line[self.__AMOUNT_PURCHASED])),
                        crypto_in=line[self.__AMOUNT_PURCHASED],
                        crypto_fee=None,
                        fiat_in_no_fee=line[self.__TOTAL_PRICE],
                        fiat_in_with_fee=line[self.__TOTAL_PRICE],
                        fiat_ticker="JPY",
                        notes=None,
                    )
                )

        return result



# Copyright 2022 eprbell
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from datetime import datetime, timedelta, timezone
from typing import List, Optional

from Historic_Crypto import HistoricalData
from rp2.rp2_decimal import RP2Decimal

from dali.abstract_pair_converter_plugin import AbstractPairConverterPlugin
from dali.historical_bar import HistoricalBar


class PairConverterPlugin(AbstractPairConverterPlugin):
    def name(self) -> str:
        return "Historic-Crypto"

    def cache_key(self) -> str:
        return self.name()

    def get_historic_bar_from_native_source(self, timestamp: datetime, from_asset: str, to_asset: str, exchange: str) -> Optional[HistoricalBar]:
        result: Optional[HistoricalBar] = None
        time_granularity: List[int] = [60, 300, 900, 3600, 21600, 86400]
        # Coinbase API expects UTC timestamps only, see the forum discussion here:
        # https://forums.coinbasecloud.dev/t/invalid-end-on-product-candles-endpoint/320
        utc_timestamp = timestamp.astimezone(timezone.utc)
        from_timestamp: str = utc_timestamp.strftime("%Y-%m-%d-%H-%M")
        retry_count: int = 0

        while retry_count < len(time_granularity):
            try:
                seconds = time_granularity[retry_count]
                to_timestamp: str = (utc_timestamp + timedelta(seconds=seconds)).strftime("%Y-%m-%d-%H-%M")
                historical_data = HistoricalData(f"{from_asset}-{to_asset}", seconds, from_timestamp, to_timestamp, verbose=False).retrieve_data()
                historical_data.index = historical_data.index.tz_localize("UTC")  # The returned timestamps in the index are timezone naive
                historical_data_series = historical_data.reset_index().iloc[0]
                result = HistoricalBar(
                    duration=timedelta(seconds=seconds),
                    timestamp=historical_data_series.time,
                    open=RP2Decimal(str(historical_data_series.open)),
                    high=RP2Decimal(str(historical_data_series.high)),
                    low=RP2Decimal(str(historical_data_series.low)),
                    close=RP2Decimal(str(historical_data_series.close)),
                    volume=RP2Decimal(str(historical_data_series.volume)),
                )
                break
            except ValueError:
                retry_count += 1

        return result



# Copyright 2023 jamesbaber1
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import List
import unittest

from rp2.rp2_decimal import ZERO, RP2Decimal
from rp2.plugin.country.us import US

from dali.configuration import Keyword
from dali.abstract_transaction import AbstractTransaction
from dali.in_transaction import InTransaction
from dali.out_transaction import OutTransaction
from dali.intra_transaction import IntraTransaction
from dali.plugin.input.ods.rp2_input import InputPlugin


class TestRP2InputOds(unittest.TestCase):

    transactions: List[AbstractTransaction]

    @classmethod
    def setUpClass(cls) -> None:
        plugin = InputPlugin(
            configuration_path='input/test_ods_rp2_input.ini',
            input_file='input/test_ods_rp2_input.ods',
            native_fiat="USD"
        )
        cls.transactions = plugin.load(US())

    def test_all_transactions_found(self) -> None:
        assert len(self.transactions) == 9

    def test_in_transactions(self) -> None:
        # Buy BTC transaction 1
        btc_transaction_in_1: InTransaction = self.transactions[0]  # type: ignore
        assert btc_transaction_in_1.asset == "BTC"
        assert btc_transaction_in_1.exchange == "FTX"
        assert btc_transaction_in_1.holder == "Bob"
        assert btc_transaction_in_1.timestamp == "2020-06-28 14:38:40+0000"
        assert btc_transaction_in_1.transaction_type == Keyword.BUY.value.capitalize()
        assert RP2Decimal(str(btc_transaction_in_1.spot_price)) == RP2Decimal("12000.00")
        assert RP2Decimal(str(btc_transaction_in_1.crypto_in)) == RP2Decimal("0.01")
        assert btc_transaction_in_1.crypto_fee is None
        assert RP2Decimal(str(btc_transaction_in_1.fiat_in_no_fee)) == RP2Decimal("120.00")
        assert RP2Decimal(str(btc_transaction_in_1.fiat_in_with_fee)) == RP2Decimal("130.00")
        assert RP2Decimal(str(btc_transaction_in_1.fiat_fee)) == RP2Decimal("10.00")

        # Buy BTC transaction 2
        btc_transaction_in_2: InTransaction = self.transactions[1]  # type: ignore
        assert btc_transaction_in_2.asset == "BTC"
        assert btc_transaction_in_1.exchange == "FTX"
        assert btc_transaction_in_1.holder == "Bob"
        assert btc_transaction_in_2.timestamp == "2022-01-02 18:11:09+0000"
        assert btc_transaction_in_2.transaction_type == Keyword.BUY.value.capitalize()
        assert RP2Decimal(str(btc_transaction_in_2.spot_price)) == RP2Decimal("35000.00")
        assert RP2Decimal(str(btc_transaction_in_2.crypto_in)) == RP2Decimal("0.99")
        assert btc_transaction_in_2.crypto_fee is None
        assert RP2Decimal(str(btc_transaction_in_2.fiat_in_no_fee)) == RP2Decimal("34650.00")
        assert RP2Decimal(str(btc_transaction_in_2.fiat_in_with_fee)) == RP2Decimal("35140.00")
        assert RP2Decimal(str(btc_transaction_in_2.fiat_fee)) == RP2Decimal("490.00")

        # Buy ETH transaction 1
        eth_transaction_in_1: InTransaction = self.transactions[6]  # type: ignore
        assert eth_transaction_in_1.asset == "ETH"
        assert eth_transaction_in_1.exchange == "Coinbase"
        assert eth_transaction_in_1.holder == "Bob"
        assert eth_transaction_in_1.timestamp == "2020-06-03 11:23:00+0000"
        assert eth_transaction_in_1.transaction_type == Keyword.BUY.value.capitalize()
        assert RP2Decimal(str(eth_transaction_in_1.spot_price)) == RP2Decimal("244")
        assert RP2Decimal(str(eth_transaction_in_1.crypto_in)) == RP2Decimal("10")
        assert eth_transaction_in_1.crypto_fee is None
        assert RP2Decimal(str(eth_transaction_in_1.fiat_in_no_fee)) == RP2Decimal("2440")
        assert RP2Decimal(str(eth_transaction_in_1.fiat_in_with_fee)) == RP2Decimal("2465")
        assert RP2Decimal(str(eth_transaction_in_1.fiat_fee)) == RP2Decimal("25.00")

    def test_out_transactions(self) -> None:
        # Sell BTC
        btc_transaction_out_1: OutTransaction = self.transactions[5]  # type: ignore
        assert btc_transaction_out_1.asset == "BTC"
        assert btc_transaction_out_1.timestamp == "2022-02-09 11:45:34-0800"
        assert btc_transaction_out_1.transaction_type == Keyword.GIFT.value.capitalize()
        assert RP2Decimal(str(btc_transaction_out_1.spot_price)) == RP2Decimal("44628.22")
        assert RP2Decimal(str(btc_transaction_out_1.crypto_out_no_fee)) == RP2Decimal("0.02")
        assert RP2Decimal(str(btc_transaction_out_1.crypto_out_with_fee)) == RP2Decimal("0.02")
        assert RP2Decimal(str(btc_transaction_out_1.crypto_fee)) == ZERO
        assert btc_transaction_out_1.fiat_fee is None

    def test_intra_transactions(self) -> None:
        # BTC transfer 1
        btc_transaction_intra_1: IntraTransaction = self.transactions[2]  # type: ignore
        assert btc_transaction_intra_1.asset == "BTC"
        assert btc_transaction_intra_1.timestamp == "2022-01-14 03:23:38-0800"
        assert RP2Decimal(str(btc_transaction_intra_1.spot_price)) == RP2Decimal("41952.11")
        assert btc_transaction_intra_1.from_exchange == "FTX"
        assert btc_transaction_intra_1.to_exchange == "Green Trezor"
        assert btc_transaction_intra_1.from_holder == "Bob"
        assert btc_transaction_intra_1.to_holder == "Bob"
        assert RP2Decimal(str(btc_transaction_intra_1.crypto_sent)) == RP2Decimal("0.2002")
        assert RP2Decimal(str(btc_transaction_intra_1.crypto_received)) == RP2Decimal("0.20")

        # BTC transfer 2
        btc_transaction_intra_2: IntraTransaction = self.transactions[3]  # type: ignore
        assert btc_transaction_intra_2.asset == "BTC"
        assert btc_transaction_intra_2.timestamp == "2022-01-18 15:27:18+0000"
        assert RP2Decimal(str(btc_transaction_intra_2.spot_price)) == RP2Decimal("41736.81")
        assert btc_transaction_intra_2.from_exchange == "Green Trezor"
        assert btc_transaction_intra_2.to_exchange == "FTX"
        assert btc_transaction_intra_2.from_holder == "Bob"
        assert btc_transaction_intra_2.to_holder == "Bob"
        assert RP2Decimal(str(btc_transaction_intra_2.crypto_sent)) == RP2Decimal("0.1001")
        assert RP2Decimal(str(btc_transaction_intra_2.crypto_received)) == RP2Decimal("0.10")

        # BTC transfer 3
        btc_transaction_intra_3: IntraTransaction = self.transactions[4]  # type: ignore
        assert btc_transaction_intra_3.asset == "BTC"
        assert btc_transaction_intra_3.timestamp == "2022-01-25 02:58:40-0800"
        assert RP2Decimal(str(btc_transaction_intra_3.spot_price)) == RP2Decimal("36463.4")
        assert btc_transaction_intra_3.from_exchange == "Green Trezor"
        assert btc_transaction_intra_3.to_exchange == "My Fave Trezor"
        assert btc_transaction_intra_3.from_holder == "Bob"
        assert btc_transaction_intra_3.to_holder == "Alice"
        assert RP2Decimal(str(btc_transaction_intra_3.crypto_sent)) == RP2Decimal("0.0503")
        assert RP2Decimal(str(btc_transaction_intra_3.crypto_received)) == RP2Decimal("0.05")
